# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''操作AndroidManifest.xml文件
'''


from qt4a.apktool.__init__ import APKError
from qt4a.apktool.axml import AXMLFile
from qt4a.apktool.apkfile import APKFile


class AndroidManifest(object):
    '''
    '''
    def __init__(self, apk_fp_or_path):
        if isinstance(apk_fp_or_path, APKFile):
            self._apk_file = apk_fp_or_path
        else:
            self._apk_file = APKFile(apk_fp_or_path)
        self._dom = self.get_dom()
        
    def get_dom(self):
        '''获取AndroidManifest.xml文件的dom树
        '''
        fp = self._apk_file.get_file('AndroidManifest.xml')
        if fp == None: raise RuntimeError('apk %s is invalid' % self._apk_path)
        axf = AXMLFile(fp)
        return axf.to_xml()
    
    @property
    def package_name(self):
        '''包名
        '''
        return self._dom.getElementsByTagName('manifest')[0].getAttribute('package')
    
    @property
    def version_code(self):
        '''versionCode
        '''
        return int(self._dom.getElementsByTagName('manifest')[0].getAttribute('android:versionCode'))
    
    @property
    def version_name(self):
        '''versionName
        '''
        return self._dom.getElementsByTagName('manifest')[0].getAttribute('android:versionName')
    
    @property
    def min_sdk_version(self):
        '''最低SDK版本
        '''
        return int(self._dom.getElementsByTagName('uses-sdk')[0].getAttribute('android:minSdkVersion'))
    
    @property
    def target_sdk_version(self):
        '''目标SDK版本
        '''
        return int(self._dom.getElementsByTagName('uses-sdk')[0].getAttribute('android:targetSdkVersion'))
        
    @property
    def application_name(self):
        '''
        '''
        return self._dom.getElementsByTagName('application')[0].getAttribute('android:name')
    
    @property
    def debuggable(self):
        '''是否是调试版本
        '''
        return self._dom.getElementsByTagName('application')[0].getAttribute('android:debuggable')
    
    @debuggable.setter
    def debuggable(self, value):
        '''设置调试标志位
        '''
        self._dom.getElementsByTagName('application')[0].setAttribute('android:debuggable', 'true' if value else 'false')
        
    @property
    def start_activity(self):
        '''启动Activity
        '''
        for activity in self._dom.getElementsByTagName('activity'):
            for filter in activity.getElementsByTagName('action'):
                if filter.getAttribute('android:name') == 'android.intent.action.MAIN':
                    activity_name = activity.getAttribute('android:name')
                    if activity_name.startswith('.'):
                        activity_name = self.package_name + activity_name
                    return activity_name
        raise APKError('apk %s do not have start activity' % self._apk_path)
    
    def get_process_list(self):
        '''获取进程列表
        '''
        process_list = []
        for tag_name in ('activity', 'service', 'provider', 'receiver'):
            for it in self._dom.getElementsByTagName(tag_name):
                process = it.getAttribute('android:process')
                if process and not process in process_list: 
                    process_list.append(process)
        return process_list
    
    def add_provider(self, name, authorities, process=None):
        '''添加ContentProvider
        '''
        elem = self._dom.createElement('provider')
        elem.setAttribute('android:name', name)
        elem.setAttribute('android:authorities', authorities)
        if process: elem.setAttribute('android:process', process)
        self._dom.getElementsByTagName('application')[0].appendChild(elem)
    
    def add_activity(self, activity, exported="true", process=None):
        elem = self._dom.createElement('activity')
        elem.setAttribute('android:name', activity)
        elem.setAttribute('android:exported', exported)
        if process: elem.setAttribute('android:process', process)
        self._dom.getElementsByTagName('application')[0].appendChild(elem)
        
    def save(self):
        '''保存修改到apk文件
        '''
        axml = AXMLFile.from_xml(self._dom)
        self._apk_file.add_file('AndroidManifest.xml', axml.serialize())
        
if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''定义Android控件
'''

import six
from io import BytesIO
import os
import tempfile
import time

from testbase.util import LazyInit, Timeout
from tuia.exceptions import ControlNotFoundError

from qt4a.androiddriver.androiddriver import AndroidDriver
from qt4a.androiddriver.util import Deprecated, get_intersection, logger

def lazy_init(func):
    '''调用到此函数时进入延迟初始化逻辑
    '''
    def wrap_func(self, *args, **kwds):
        self._lazy_obj._app  # 调到窗口类的延迟初始化函数
        return func(self, *args, **kwds)
    return wrap_func

class Window(object):
    '''控件容器基类
    '''
    Process = ''  # 所在的进程名，不填表示是主进程
    Activity = ''  # 所在的Activity

    def __init__(self, app_or_driver, wait_activity=True, **kwds):
        self._locators = {}
        self._device = None
        self._app = None
        if hasattr(app_or_driver, '_drivers'):
            # AndroidApp类型
            self._driver = app_or_driver.get_driver(self.Process)
            self._device = app_or_driver.device
            self._app = app_or_driver
        elif hasattr(app_or_driver, '_driver'):
            # Window类型
            self._app = app_or_driver._app
            if self._app:
                self._driver = app_or_driver._app.get_driver(self.Process)
            else:
                self._driver = app_or_driver._driver
            self._device = app_or_driver.device

        else:
            # AndroidDriver类型，一般用于调试
            self._driver = app_or_driver
        
        if not isinstance(self._driver, AndroidDriver):
            raise RuntimeError('参数类型错误：%s' % self._driver.__class__)
        self._wait_activity = wait_activity
        if wait_activity:
            self._wait_timeout = kwds.get('wait_timeout')
            self._wait_interval = kwds.get('wait_interval')
        self._lazy_obj = LazyInit(self, '_lazy_obj', self._init_window)

    def _init_window(self):
        '''延迟初始化时执行
        '''
        self._lazy_obj = self  # 避免多次调用初始化函数
        if self._wait_activity == True:
            wait_timeout = self._wait_timeout
            if not wait_timeout: wait_timeout = 12
            wait_interval = self._wait_interval
            if not wait_interval: wait_interval = 0.5
            self.wait_for_exist(wait_timeout, wait_interval)
        self.post_init()
        return self
    
    def post_init(self):
        '''窗口类自定义的初始化逻辑
        '''
        pass
    
    @property
    def Controls(self):
        '''返回控件集合。使用如foo.Controls['最小化按钮']的形式获取控件
        '''
        return self
    
    def hasControlKey(self, control_key):
        '''是否包含控件control_key
        
        :rtype: boolean
        '''
        return (control_key in self._locators)
    
    @property
    def device(self):
        return self._device

    def __findctrl_recur(self, ctrlkey):
        if not (ctrlkey in self._locators.keys()):
            raise NameError("%s没有名为'%s'的子控件！" % (type(self), ctrlkey))
        params = self._locators[ctrlkey].copy()
        ctrltype = params['type']
        del params['type']
        for key in params:
            value = params[key]
            if isinstance(value, six.string_types) and value.startswith('@'):
                params[key] = self[value[1:]]  # 使用缓存
        if issubclass(ctrltype, View):
            return ctrltype(**params)
        else:
            root = params.pop('root', None)
            if root == None: root = self
            params.pop('driver', None)  # delete unexpected param
            params.pop('activity', None)
            return ctrltype(root, ctrlkey, **params)
            
    @lazy_init
    def __getitem__(self, index):
        '''获取index指定控件
        
        :type index: string
        :param index: 控件索引，如'查找按钮'  
        '''
        if not (index in self._locators.keys()):
            raise NameError("%s没有名为'%s'的子控件！" % (type(self), index))
        if not '_instance' in self._locators[index]:
            instance = self.__findctrl_recur(index)
            self._locators[index]['_instance'] = instance
        return self._locators[index]['_instance']
    
    @Deprecated('update_locator')
    def updateLocator(self, locators):
        self.update_locator(locators)
    
    def update_locator(self, locators):
        '''更新控件定义
        '''
        for key in locators.keys():
            locators[key]['driver'] = self._driver  # 保存driver实例
            locators[key]['activity'] = self.__class__.Activity
        self._locators.update(locators)
        
    def wait_for_exist(self, timeout=10, interval=0.5):
        '''等待窗口出现
        '''
        import re
        time0 = time.time()
        current_activity = ''
        if self.Activity == '': return True
        pattern = re.compile(self.Activity)
        while time.time() - time0 < timeout:
            current_activity = self.device.current_activity
            if current_activity == self.Activity:
                return True
            if current_activity and pattern.match(current_activity):
                return True
            time.sleep(interval)
        raise ControlNotFoundError('窗口：%s 未找到，当前窗口为：%s' % (self.__class__.Activity, current_activity))
    
    @property
    @lazy_init
    def rect(self):
        '''窗口区域
        '''
        root = self._driver.get_control(self.Activity, None, [])
        if root == 0:
            raise RuntimeError('查找根节点失败')
        rect = self._driver.get_control_rect(root)
        return rect['Left'], rect['Top'], rect['Width'], rect['Height']

    @lazy_init
    def close(self):
        '''关闭窗口
        '''
        if self.Activity == '':
            self._driver.send_key('{BACK}')
        else:
            if not self._driver.close_activity(self.Activity):
                self._driver.send_key('{BACK}')

        timeout = 3
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self._driver._device.get_current_activity() != self.Activity: return
            time.sleep(0.5)
    
    def get_metis_view(self):
        '''返回MetisView
        '''
        from qt4a.mtcontrols import MetisView
        return MetisView(self)
    
class Gravity(object):
    '''用于控制控件靠左、居中等位置的属性
    '''
    NO_GRAVITY = 0x0000
    AXIS_SPECIFIED = 0x0001
    AXIS_PULL_BEFORE = 0x0002
    AXIS_PULL_AFTER = 0x0004
    AXIS_CLIP = 0x0008
    AXIS_X_SHIFT = 0
    AXIS_Y_SHIFT = 4

    TOP = (AXIS_PULL_BEFORE | AXIS_SPECIFIED) << AXIS_Y_SHIFT
    BOTTOM = (AXIS_PULL_AFTER | AXIS_SPECIFIED) << AXIS_Y_SHIFT
    LEFT = (AXIS_PULL_BEFORE | AXIS_SPECIFIED) << AXIS_X_SHIFT
    RIGHT = (AXIS_PULL_AFTER | AXIS_SPECIFIED) << AXIS_X_SHIFT
    CENTER_VERTICAL = AXIS_SPECIFIED << AXIS_Y_SHIFT
    FILL_VERTICAL = TOP | BOTTOM
    CENTER_HORIZONTAL = AXIS_SPECIFIED << AXIS_X_SHIFT
    FILL_HORIZONTAL = LEFT | RIGHT
    CENTER = CENTER_VERTICAL | CENTER_HORIZONTAL
    FILL = FILL_VERTICAL | FILL_HORIZONTAL
    CLIP_VERTICAL = AXIS_CLIP << AXIS_Y_SHIFT
    CLIP_HORIZONTAL = AXIS_CLIP << AXIS_X_SHIFT
    RELATIVE_LAYOUT_DIRECTION = 0x00800000
    HORIZONTAL_GRAVITY_MASK = (AXIS_SPECIFIED | AXIS_PULL_BEFORE | AXIS_PULL_AFTER) << AXIS_X_SHIFT
    VERTICAL_GRAVITY_MASK = (AXIS_SPECIFIED | AXIS_PULL_BEFORE | AXIS_PULL_AFTER) << AXIS_Y_SHIFT
    DISPLAY_CLIP_VERTICAL = 0x10000000
    DISPLAY_CLIP_HORIZONTAL = 0x01000000
    START = RELATIVE_LAYOUT_DIRECTION | LEFT
    END = RELATIVE_LAYOUT_DIRECTION | RIGHT
    RELATIVE_HORIZONTAL_GRAVITY_MASK = START | END

    def __init__(self, grav):
        self._gravity = grav

    @property
    def left(self):
        raise NotImplementedError()

    @property
    def right(self):
        raise NotImplementedError()

    @property
    def center(self):
        raise NotImplementedError()


def func_wrap(func):
    '''用于方法包装，主要作用是发现控件失效时能够更新控件
    '''
    def _func(*args, **kwargs):
        from qt4a.androiddriver.util import logger, ControlExpiredError
        try:
            return func(*args, **kwargs)
        except ControlExpiredError as e:
            view = args[0]
            try:
                view._update()
            except RuntimeError:
                logger.exception('update control failed')
                raise e  # 更新失败时直接抛出控件失效异常
            return func(*args, **kwargs)
    return _func

class VerticalSwipe(object):
    '''纵向滑动包装类
    '''
    def __init__(self, view):
        self._view = view
        self._is_class = isinstance(view, type)
        
    def __call__(self, *args, **kwds):
        if not self._is_class:
            raise RuntimeError('对象不支持__call__')
        view = self._view(*args, **kwds)
        return VerticalSwipe(view)
    
    def __str__(self):
        if not self._is_class:
            return '<VerticalSwipe_%s object at 0x%.8X>' % (self._view.__class__.__name__, id(self))
        else:
            return '<class \'VerticalSwipe_%s\'>' % self._view.__name__
        
    @property
    def rect(self):
        '''left, top, width, height
        '''
        self_rect = self._view.rect
        root = self._view._root

        root_list = []
        while not isinstance(root, Window):
            if isinstance(root, ScrollView):
                root_list.append(root)
            root = root._root
            
        if len(root_list) == 0: return self_rect
        root_rect = root_list[-1].rect  # 获取可滚动的最顶层根
        left = self_rect[0] if self_rect[0] > root_rect[0] else root_rect[0]
        top = self_rect[1] if self_rect[1] > root_rect[1] else root_rect[1]
        width = self_rect[0] + self_rect[2] - left if self_rect[0] + self_rect[2] < root_rect[0] + root_rect[2] else root_rect[0] + root_rect[2] - left
        height = self_rect[1] + self_rect[3] - top if self_rect[1] + self_rect[3] < root_rect[1] + root_rect[3] else root_rect[1] + root_rect[3] - top
        return left, top, width, height  # 取交集
        
    def swipe_up(self):
        '''向上滑动
        '''
        rect = self.rect
        x1 = x2 = rect[0] + rect[2] // 2
        y1 = rect[1] + rect[3] * 3 // 4
        y2 = rect[1] + rect[3] // 4
        self._view._driver.drag(x1, y1, x2, y2)
    
    def swipe_down(self):
        '''向下滑动
        '''
        rect = self.rect
        x1 = x2 = rect[0] + rect[2] // 2
        y1 = rect[1] + rect[3] // 4
        y2 = rect[1] + rect[3] * 3 // 4
        self._view._driver.drag(x1, y1, x2, y2)
    
    def __getattr__(self, attr):
        return getattr(self._view, attr)
    
class View(object):
    '''控件基类
    '''
    def __init__(self, activity, root, driver, locator=None, hashcode=0):
        self._root = root
        if locator != None:
            self._qpath = locator
            self._locator = locator._parsed_qpath  # 解析后的list
        else:
            if hashcode == 0:
                raise RuntimeError('参数错误')
            self._locator = []
        self._driver = driver
        self._activity = activity
        self._hashcode = 0  # 每个Java对象实例的hashcode（根据内存地址计算出来）
        if hashcode != 0:
            self._hashcode = hashcode
        self._need_convert_qpath = True  # 是否需要转换QPath的ID为整型ID
        
    def __eq__(self, other):
        '''根据hashcode判断两个控件是否相同
        '''
        if other == None: return False
        if self.hashcode == 0 or other.hashcode == 0:
            raise RuntimeError('hashcode错误')
        return self.hashcode == other.hashcode

    def __str__(self):
        return '<%s(id=0x%X) at 0x%X>' % (self.__class__.__name__, self.hashcode & 0xFFFFFFFF, id(self))
    
    def _handle_qpath(self, qpath):
        '''对qpath进行处理，返回处理过的QPath列表
        '''
        # 将字符串ID改为整型ID
        import copy, itertools
        qpath = copy.deepcopy(qpath)  # 防止qpath被修改
        
        result = []  # 返回的是可能的qpath的列表
        for item in qpath:
            item_list = []
            if 'Id' in item:
                # TODO: 如果有两级ID，其中一级是数字形式的
                id_list = self.container._app._get_view_id(item['Id'][1])  # 可能会有多个整型ID
                if id_list == None: 
                    item_list.append(item)
                    result.append(item_list)
                    continue
                if len(id_list) == 1:
                    item['Id'][1] = str(id_list[0])
                    item_list.append(item)
                else:
                    # 存在多个id
                    for _id in id_list:
                        item['Id'][1] = str(_id)
                        item_list.append(copy.deepcopy(item))
            else:
                item_list.append(item)
            result.append(item_list)
            
        qpath_list = []
        for qpath in itertools.product(*result):
            qpath_list.append(qpath)
        return qpath_list
    
    @property
    def container(self):
        '''获取所在容器类
        '''
        root = self._root
        if isinstance(root, Window):
            return root
        else:
            return root.container

    def _get_hashcode(self, parent=0, locator=None):
        '''找到控件的hashcode值
        '''
        if locator: self._locator = locator
        if not self._locator: raise RuntimeError('控件定位信息缺失')
        if parent == 0 and isinstance(self._root, View):
            # 存在父节点
            if self._root._hashcode != 0:
                parent = self._root._hashcode
            else:
                parent = self._root._get_hashcode()
                if parent == 0: return 0  # 父节点不存在
                self._root._hashcode = parent
        
        idx = 0
        if isinstance(self._locator, list):
            # 先使用原始QPath查询一次
            result = self._driver.get_control(self._activity, parent, self._locator)
            if result != 0: 
                if self._need_convert_qpath: self._need_convert_qpath = False  # 不再需要转换QPath
                return result  # TODO: 处理"控件索引超出范围"情况
            if not self._need_convert_qpath: return 0  # 避免控件消失后还要再去获取整型ID
            if not self.container._app._is_use_int_view_id(): return 0  # 不使用整型ID
            
            import re
            pattern = re.compile('^0x[0-9a-fA-F]+$')
            
            need_convert = False
            for item in self._locator:
                if 'Id' in item:
                    if item['Id'][1].isdigit(): continue
                    if pattern.match(item['Id'][1]): continue  # 16进制的
                    need_convert = True
                    break
            if not need_convert: return 0  # 已经是数字类型的ID了
            
            # 生成整型ID的QPath
            qpath_list = self._handle_qpath(self._locator)  
            if len(qpath_list) == 1:
                self._locator = (self._locator, qpath_list[0])
            elif len(qpath_list) > 1:
                # 有多个QPath
                qpath_list.insert(0, self._locator)
                self._locator = tuple(qpath_list)  # tuple类型
            else:
                raise RuntimeError('QPath解析错误：%s' % self._locator)
            idx = 1  # 第一个不需要检查了
            
        for loc in self._locator[idx:]:
            result = self._driver.get_control(self._activity, parent, loc)
            if result != 0: 
                self._locator = list(loc)  # 如果找到,认为这是正确的QPath,以后只使用该QPath进行查找
                self._need_convert_qpath = False
                return result  # 找到即返回
        return 0
            
    @property
    def hashcode(self):
        '''控件唯一标识，只有真正访问控件信息时才会去获取该标识
        '''
        if self._hashcode == 0:
            # 尚未获取控件的hashcode
            timeout = 10  # 查找控件的超时时间
            interval = 0.1
            time0 = time.time()
            while time.time() - time0 < timeout:
                self._hashcode = self._get_hashcode()
                if self._hashcode == 0:
                    # 等待一段时间再查找
                    time.sleep(interval)
                else:
                    break
            if self._hashcode == 0:
                root = self._root
                root_list = []
                while isinstance(root, View):
                    root_list.append(root)
                    root = root._root
                qpath = self._qpath
                locator = self._locator
                parent = self._root
                if len(root_list) > 0:
                    for i in range(len(root_list) - 1, -1, -1):
                        if root_list[i]._hashcode == 0:
                            qpath = root_list[i]._qpath
                            locator = root_list[i]._locator
                            parent = root_list[i]._root
                            break
                parent = parent._hashcode if isinstance(parent, View) else 0
                pos = self._driver.get_control(self._activity, parent, locator, True)
                qpath = str(qpath)
                split_char = qpath[0]
                qpath_list = qpath[1:].split(split_char)
                err_qpath = split_char.join(qpath_list[pos:])
                if err_qpath: err_qpath = split_char + err_qpath  # 补上前面的分隔符
                err_msg = '进程【%s】中未找到控件：%s \n未找到部分路径为：【%s】' % (self._driver._process_name, qpath, err_qpath)
                current_activity = self.container.device.get_current_activity()
                view_tree = self._driver.get_control_tree(current_activity)
                logger.info('Current acvitiy %s control tree：\n%s' % (current_activity.encode('utf8'), view_tree.encode('utf8')))
                raise ControlNotFoundError(err_msg)
        return self._hashcode

    def _update(self):
        '''更新hashcode
        '''
        if isinstance(self._root, View):
            self._root._update()
        self._hashcode = 0
        return self.hashcode

    @property
    @func_wrap
    def parent(self):
        '''获取父控件
        '''
        if not hasattr(self, '_parent'):
            self._parent = self._driver.get_parent(self.hashcode)
        return View(hashcode=self._parent, activity=self._activity, root=self.container, driver=self._driver)

    @property
    @func_wrap
    def children(self):
        '''获取子节点列表
        '''
        children = self._driver.get_children(self.hashcode)
        return [View(hashcode=child, root=self, activity=self._activity, driver=self._driver) for child in children]

    @property
    @func_wrap
    def rect(self):
        '''left, top, width, height
        '''
        get_rect_interval = 0.1  # 防止频繁获取控件坐标信息
        if not hasattr(self, '_last_get_rect_time') or (time.time() - self._last_get_rect_time) > get_rect_interval:
            rect = self._driver.get_control_rect(self.hashcode)
            result = rect['Left'], rect['Top'], rect['Width'], rect['Height']
            self._last_get_rect_time = time.time()
            self._last_rect = result
        return self._last_rect

    @property
    @func_wrap
    def visible(self):
        '''是否可见
        '''
        return self._driver.get_control_visibility(self.hashcode)

    @property
    @func_wrap
    def _clickable(self):
        '''是否可点击
        '''
        CLICKABLE = 0x00004000
        flags = self._driver.get_object_field_value(self.hashcode, 'mViewFlags')
        return int(flags) & CLICKABLE == CLICKABLE

    @property
    def clickable(self):
        '''是否可点击
        to be deleted
        '''
        return self._clickable

    @property
    @func_wrap
    def enabled(self):
        '''是否可用
        '''
        ENABLE = 0x00000000
        ENABLED_MASK = 0x00000020
        flags = int(self._driver.get_object_field_value(self.hashcode, 'mViewFlags'))
        return flags & ENABLED_MASK == ENABLE

    @property
    def content_desc(self):
        '''控件描述
        '''
        return self._driver.get_object_field_value(self.hashcode, 'mContentDescription')
    
    def exist(self):
        '''判断控件是否存在
        '''
        from qt4a.androiddriver.util import ControlExpiredError, logger
        try:
            self._hashcode = self._get_hashcode()
            return self._hashcode != 0  # 由于Java端使用int型存储，因此可能为负数
        except ControlExpiredError:
            return False
        except RuntimeError:
            logger.exception('获取控件失败')  # 如果没有获取到控件ID对应的整型ID，认为控件不存在
            return False
        
    def wait_for_exist(self, timeout=10, interval=0.1):
        '''等待控件出现
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.exist(): return True
            time.sleep(interval)
        raise ControlNotFoundError('控件：%s 未找到' % self._qpath)

    def wait_for_visible(self, timeout=10, interval=0.2):
        '''等待控件可见
        
        :param timeout:  超时时间，单位：秒
        :type  timeout:  int/float
        :param interval: 重试间隔时间，单位：秒
        :type  interval: int/float
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self.visible: return
            time.sleep(interval)
        raise RuntimeError('Control is not visible in %s seconds' % timeout)
    
    def wait_for_invisible(self, timeout=10, interval=0.2):
        '''等待控件不可见
        
        :param timeout:  超时时间，单位：秒
        :type  timeout:  int/float
        :param interval: 重试间隔时间，单位：秒
        :type  interval: int/float
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            if not self.visible: return
            time.sleep(interval)
        raise RuntimeError('Control is not invisible in %s seconds' % timeout)
    
    def _point_in_view(self, x, y):
        '''判断点(x, y)是否在当前View可视范围内
        '''
        view_rect = self.rect
        if x < view_rect[0] or x >= view_rect[0] + view_rect[2]: return False
        if y < view_rect[1] or y >= view_rect[1] + view_rect[3]: return False
        return True
    
    def _get_scroll_root(self):
        '''获取当前控件可滚动区域的根控件
        '''
        root = self._root
        if isinstance(root, ListItem):
            root = root._root  # 获取ListView视图

        root_list = []
        while not isinstance(root, Window):
            if isinstance(root, (ScrollView, ViewPager)):
                root_list.append(root)
            root = root._root
        if root_list == []: return None
        return root_list[-1]  # 获取可滚动的最顶层根
        
    def scroll_to_visible(self, root_rect=None, self_rect=None):
        '''滚动到可视区域
        '''
        root = self._get_scroll_root()
        if root == None: return 0, 0  # 不需要滚动
        
        if not self_rect: self_rect = self.rect
        if not root_rect: root_rect = root.rect
        visible_rect = get_intersection(self_rect, root_rect)
        min_clickable_area = (20, 10)  # 最小可点击区域的大小
        _, screen_height = self.container.device.screen_size
        if visible_rect[0] >= 0 and visible_rect[1] >= 0 and visible_rect[2] >= min_clickable_area[0] and visible_rect[3] >= min_clickable_area[1] and screen_height - (visible_rect[1] + visible_rect[3]) >= 50: 
            logger.info('intersect area is %s' % (visible_rect,))
            return 0, 0
        
        if self_rect[2] > root_rect[2] or self_rect[3] > root_rect[3]:
            x_visible = True
            y_visible = True
            if self_rect[2] >= root_rect[2]:
                # 控件宽度大于容器宽度
                max_left = self_rect[0] if self_rect[0] > root_rect[0] else root_rect[0]
                min_right = self_rect[0] + self_rect[2] if self_rect[0] + self_rect[2] < root_rect[0] + root_rect[2] else root_rect[0] + root_rect[2]
                visible_width = min_right - max_left
                if visible_width < root_rect[2] // 2:
                    # 由于此时控件不可能完全可见，因此认为一半可见即为可见
                    x_visible = False
            if self_rect[3] >= root_rect[3]:
                # 控件高度大于容器高度
                max_top = self_rect[1] if self_rect[1] > root_rect[1] else root_rect[1]
                min_bottom = self_rect[1] + self_rect[3] if self_rect[1] + self_rect[3] < root_rect[1] + root_rect[3] else root_rect[1] + root_rect[3]
                visible_height = min_bottom - max_top
                if visible_height < root_rect[3] // 2:
                    y_visible = False

            if x_visible and y_visible: return 0, 0
            
        # print (self_rect, root_rect)

        if isinstance(root, ViewPager):
            offset = self_rect[0] - root_rect[0]
            count = offset // root_rect[2]
            root.scroll(count, root_rect)
            return root_rect[2] * count, 0

        x_offset = y_offset = 0
        if self_rect[0] < root_rect[0]:
            # 需要向右滑动
            x_offset = self_rect[0] - root_rect[0]
        elif self_rect[0] + self_rect[2] > root_rect[0] + root_rect[2]:
            # 理论上不会出现左右边界都超出root边界范围的情况
            x_offset = (self_rect[0] + self_rect[2]) - (root_rect[0] + root_rect[2])
        if self_rect[1] < root_rect[1]:
            # 需要向下滑动
            y_offset = self_rect[1] - root_rect[1]
        elif self_rect[1] + self_rect[3] > root_rect[1] + root_rect[3]:
            # print (self_rect[1], self_rect[3], root_rect[1], root_rect[3])
            y_offset = (self_rect[1] + self_rect[3]) - (root_rect[1] + root_rect[3])

        if abs(x_offset) > 5 or abs(y_offset) > 5:
            # 低于5个像素不滚动
            # print (x_offset, y_offset)
            root.scroll(x_offset, y_offset)
            time.sleep(0.5)  # 测试发现，滑动后等待时间过短就点击会无效
        return x_offset, y_offset

    def _pre_click(self, x_offset=None, y_offset=None):
        '''点击前的处理
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        
        :return 要点击的坐标
        '''

        visible = False
        rect_valid = False
        old_rect = None
        enable = False
        
        screen_width, screen_height = self.container.device.screen_size
        self.hashcode  # 确保控件存在
        root = self._get_scroll_root()
        root_rect = root.rect if root != None else [0, 0, screen_width, screen_height]
        if root != None:
            timeout = 2
            time0 = time.time()
            while time.time() - time0 < timeout:
                if root_rect[2] == 0 or root_rect[3] == 0 or root_rect[0] < 0 or root_rect[1] < 0 or \
                root_rect[0] + root_rect[2] > screen_width or root_rect[1] + root_rect[3] > screen_height:
                    time.sleep(0.1)
                    root_rect = root.rect
                else:
                    break
            else:
                logger.warn('invalid root rect: %s' % (root_rect,))
                
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            # 尝试操作，会出现由于控件尚未初始化完成导致获取的rect不正确的情况
            if not visible:
                if not self.visible:
                    time.sleep(0.1)
                    continue
                else:
                    visible = True

            if not enable:
                # 检查是否可用
                enable = self.enabled
                if not enable:
                    time.sleep(0.1)
                    continue
            
            if not rect_valid:
                rect = self.rect
                if rect[2] == 0 or rect[3] == 0:  # 进行点击操作的控件长宽不可能为0
                    logger.debug('control %s width or height is 0 [%s]' % (self._hashcode, rect))
                    time.sleep(0.2)  # 间隔时间延长为0.2S
                    continue
                
                if not old_rect or rect != old_rect:
                    # 等待控件坐标稳定
                    old_rect = rect
                    logger.debug('wait for control %s stable' % self._hashcode)
                    time.sleep(0.2)
                    continue
                
                is_scroll = False  # 控件是否滚动
                if root != None:
                    for _ in range(5):
                        x_offset1, y_offset1 = self.scroll_to_visible(root_rect, rect)  # 自动滚动到可视区域
                        if abs(x_offset1) <= 5 and abs(y_offset1) <= 5:
                            break
                        else:
                            is_scroll = True
                            rect = self.rect  # 重新获取滚动后的坐标
                
                visible_rect = get_intersection(rect, root_rect)
                if is_scroll:
                    old_rect = visible_rect
                    time.sleep(0.2)
                else:
                    # 不需要再判断控件是否稳定
                    old_rect = None
                    rect = visible_rect
                rect_valid = True
            
            if old_rect != None:
                rect = get_intersection(self.rect, root_rect)
                if old_rect != rect:
                    # 防止有些控件加载后出现位移
                    old_rect = rect
                    time.sleep(0.2)
                    continue
            
            x = (rect[0] + rect[2] // 2)
            y = (rect[1] + rect[3] // 2)
            if x_offset: x = rect[0] + x_offset
            if y_offset: y = rect[1] + y_offset
            return x, y
        else:
            current_activity = self.container.device.get_current_activity()
            if current_activity != self._activity:
                import re
                ret = re.compile(self._activity)
                if not ret.match(current_activity): raise RuntimeError("当前Activity为%s，不是控件所属的Activity" % current_activity)
            if not visible: raise RuntimeError('控件不可见')
            if not enable: raise RuntimeError('控件不可用')
            if not rect_valid: raise RuntimeError('控件区域错误')
            raise RuntimeError('未知错误')
    
    def _click(self, click_time, x=None, y=None, check_ret=True):
        '''具有重试逻辑的点击
        
        :param click_time: 按住的时长，单位为秒
        :type click_time:  int/float
        :param x: 点击的横坐标
        :type x:  int/float
        :param y: 点击的纵坐标
        :type y:  int/float
        :param check_ret: 是否检查点击成功
        :type check_ret:  bool
        '''
        if not x or not y:
            rect = self.rect
            if not x: x = rect[0] + rect[2] // 2
            if not y: y = rect[1] + rect[3] // 2
        for _ in range(10):
            ret = self._driver.click(self.hashcode, x, y, click_time)
            if not check_ret or ret: return True
            time.sleep(0.5)
        raise RuntimeError('click (%d, %d) 失败' % (x, y))
    
    @func_wrap
    def click(self, x_offset=0, y_offset=0):
        '''单击
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        x, y = self._pre_click(x_offset, y_offset)
        self._click(0, x, y)
    
    def double_click(self, x_offset=0, y_offset=0):
        '''双击
        
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        x, y = self._pre_click(x_offset, y_offset)
        self._click(0, x, y)
        self._click(0, x, y)
        
    def multiple_click(self, count=3, x_offset=0, y_offset=0):
        '''多次点击
        
        :param count: 点击多少次
        :type count:  int
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        '''
        if count < 1: return
        x, y = self._pre_click(x_offset, y_offset)
        for _ in range(count):
            self._click(0, x, y)
    
    @func_wrap
    def long_click(self, duration=1, x_offset=0, y_offset=0, sync=True):
        '''长按
        
        :param duration: 按住时长，单位为秒
        :type duration:  int
        :param x_offset: 距离控件区域左上角的横向偏移。
        :type x_offset:  int或float
        :param y_offset: 距离控件区域左上角的纵向偏移。
        :type y_offset:  int或float
        :param sync:     是否是同步调用，为True表示等到长按结束才返回，False表示立即返回
        :type sync:      bool
        '''
        x, y = self._pre_click(x_offset, y_offset)
        if sync:
            self._click(duration, x, y, False)  # 长按不检查回调
        else:
            import threading
            # 按下后立即返回
            self._driver.drag(x, y, x, y, 0, 0, True, False)
            def _delay_func():
                time.sleep(duration)
                self._driver.drag(x, y, x, y, 0, 0, False, True)
            t = threading.Thread(target=_delay_func, args=())
            t.start()
            
    def wait_for_value(self, prop_name, prop_value, timeout=10, interval=0.5, regularMatch=False):
        """等待控件属性值出现, 如果属性为字符串类型，则使用正则匹配
        
        :param prop_name: 属性名字
        :param prop_value: 等待出现的属性值
        :param timeout: 超时秒数, 默认为10
        :param interval: 等待间隔，默认为0.5
        :param regularMatch: 参数 property_name和waited_value是否采用正则表达式的比较。默认为不采用（False）正则，而是采用恒等比较。
        """
        Timeout(timeout, interval).waitObjectProperty(self, prop_name, prop_value, regularMatch)
    
    def swipe(self, direct):
        '''滑动
        
        :param direct: 方向
        :type direct:  string，只能是“up”、“down”、“left”、“right”中的一个值
        '''
        rect = self.rect
        if direct == 'up':
            x1 = x2 = rect[0] + rect[2] // 2
            y1 = rect[1] + rect[3] * 2 // 3
            y2 = rect[1] + rect[3] // 3
        elif direct == 'down':
            x1 = x2 = rect[0] + rect[2] // 2
            y1 = rect[1] + rect[3] // 3
            y2 = rect[1] + rect[3] * 2 // 3
        elif direct == 'left':
            y1 = y2 = rect[1] + rect[3] // 2
            x1 = rect[0] + rect[2] * 2 // 3
            x2 = rect[0] + rect[2] // 3
        elif direct == 'right':
            y1 = y2 = rect[1] + rect[3] // 2
            x1 = rect[0] + rect[2] // 3
            x2 = rect[0] + rect[2] * 2 // 3
        else:
            raise RuntimeError('direct参数只能是：up、down、left、right中的一个')  
        self._driver.drag(x1, y1, x2, y2)
    
    def get_metis_view(self):
        '''返回MetisView
        '''
        from qt4a.mtcontrols import MetisView
        return MetisView(self)
    
class TextView(View):
    '''包含Text的View 
    '''
    @property
    @func_wrap
    def text(self):
        '''获取文本
        '''
        return self._driver.get_control_text(self.hashcode)
    
    @property
    def html_style_text(self):
        '''HTML格式文本
        '''
        return self._driver.get_control_text(self.hashcode, True)
    
    @text.setter
    def text(self, value):
        '''设置文本
        '''
        if not isinstance(value, six.string_types):
            value = str(value)
        if six.PY2 and not isinstance(value, unicode):
            try:
                value = value.decode('utf8')
            except UnicodeDecodeError:
                raise RuntimeError('参数编码错误：%r' % value)
        self.wait_for_visible()
        self.disable_soft_input()  # 赋值前关闭软键盘
        for _ in range(3):
            if self._driver.set_control_text(self.hashcode, value): return
            self.click()
        raise RuntimeError('设置控件文本失败')
    
    @property
    def text_size(self):
        '''字体大小
        '''
        ret = self._driver.call_object_method(self.hashcode, 'mTextPaint', 'getTextSize')
        return int(ret)
    
    @property
    def text_color(self):
        '''字体颜色
        '''
        import json
        ret = json.loads(self._driver.get_object_field_value(self.hashcode, 'mCurTextColor'))
        ret = int(ret)
        if ret < 0: ret += 0x100000000
        return ret
    
    @func_wrap
    def disable_soft_input(self):
        '''禁用软键盘
        '''
        self._driver.enable_soft_input(self.hashcode, False)
    
    def _get_char_rect(self, start_offset, end_offset):
        '''根据字符偏移量计算区域坐标
        
        :param start_offset: 起始字符偏移
        :type start_offset:  int
        :param end_offset:   结束字符偏移
        :type end_offset:    int
        '''
        rect = self._driver.get_text_rect(self.hashcode, start_offset, end_offset)
        return rect['Left'], rect['Top'], rect['Width'], rect['Height']
    
    @property
    def hint_text(self):
        '''空白提示文本
        '''
        return self._driver.get_object_field_value(self.hashcode, 'mHint')
    
    @property
    def image_resource_name(self):
        '''图像资源名称
        '''
        result = self._driver.get_control_image_resource(self.hashcode)
        for key in result:
            result[key] = self.container._app._get_drawable_resource_origin_name(result[key])
        return result
    
    def click_clickable_span(self):
        '''点击TextView中的ClickableSpan区域
        '''
        import re
        pattern = re.compile(r'^<p.*>(.|\n)*<span onclick>(.+)</span>(.*)</p>\s*$')
        html = self.html_style_text
        ret = pattern.match(html)
        if not ret: raise RuntimeError('HTML文本格式错误: %r' % html)
        text = self.text
        start_offset = text.find(ret.group(2))
        end_offset = start_offset + len(ret.group(2))
        rect = self._get_char_rect(start_offset + 1, end_offset)

        textview_rect = self.rect
        x1 = textview_rect[0]
        y1 = textview_rect[1]
        x2 = rect[0] + rect[2] // 2
        y2 = rect[1] + rect[3] // 2
        self.click(x2 - x1, y2 - y1)
        
class EditText(TextView):
    '''输入文本框
    '''
    def send_text(self, text):
        '''输入按键，此方法不能输入中文和大写字母
        '''
        self.click()  # 先获取焦点
        time.sleep(0.1)
        self._driver.send_keys(text)

    def click(self):
        '''click后自动关闭输入法
        '''
        super(TextView, self).click()
        self.disable_soft_input()
        time.sleep(1)

class Button(TextView):
    '''按钮类
    '''
    pass

class CompoundButton(Button):
    '''可选按钮，一般用于实现Switch
    '''
    @property
    @func_wrap
    def checked(self):
        '''是否已选
        '''
        return self._driver.get_control_checked(self.hashcode)

    @checked.setter
    def checked(self, check=True):
        '''设置是否选择
        '''
        if self.checked == check: return
        self.click()

class RadioButton(CompoundButton):
    '''单选按钮
    '''
    pass

class CheckBox(CompoundButton):
    '''选择按钮
    '''
    pass

class CheckedTextView(TextView):
    '''可选文本框
    '''
    @property
    def checked(self):
        '''是否已选
        '''
        return self._driver.get_object_field_value(self.hashcode, 'mChecked') == 'true'

class ImageView(View):
    '''显示图片控件
    '''
    @property
    @func_wrap
    def resource_name(self):
        '''当前使用的图片资源名称
        '''
        ret = self._driver.get_control_background(self.hashcode)
        if ret: return self.container._app._get_drawable_resource_origin_name(ret)
        res_name = self._driver.get_control_image_resource(self.hashcode)
        return self.container._app._get_drawable_resource_origin_name(res_name)
    
    def capture(self, save_path=''):
        '''保存图片到本地
        '''
        if not save_path:
            import tempfile, os
            tmp_path = tempfile.mkdtemp()
            save_path = os.path.join(tmp_path, 'tmp.png')
        
        for _ in range(3):
            pic_data = self._driver.capture_control(self.hashcode)
            if len(pic_data) == 0: 
                time.sleep(1)
                continue
            f = open(save_path, 'wb')
            f.write(pic_data)
            f.close()
            return save_path
        
        raise RuntimeError('获取控件截图失败')
    
    def save(self, save_path=''):
        '''保存图片到本地
        to de deleted
        '''
        return self.capture(save_path)
    
class ImageButton(ImageView):
    pass


class DropdownView(View):
    pass


# 布局类
class ViewGroup(View):
    pass

class FrameLayout(ViewGroup):
    pass

class LinearLayout(ViewGroup):
    pass

class RelativeLayout(ViewGroup):
    '''
    '''
    @property
    def gravity(self):
        '''位置属性
        '''
        return Gravity(self._driver.get_control_gravity(self.hashcode))

class ProgressBar(View):
    '''进度条
    '''
    @property
    def progress(self):
        '''进度
        '''
        return self._driver.get_control_progress(self.hashcode)

class SeekBar(ProgressBar):
    '''可修改进度的进度条
    '''
    @property
    def progress(self):
        '''进度
        '''
        return super(SeekBar, self).progress
    
    @progress.setter
    def progress(self, new_progress):
        '''设置新的进度
        
        :param new_progress: 新进度值,取值范围:0-100
        :type new_progress:  float
        '''
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            rect = self.rect
            if rect[2] == 0 or rect[3] == 0:
                time.sleep(0.2)
                continue
            break
        offset_y = rect[3] // 2
        offset_x = rect[2] * new_progress / 100.0
        self.click(offset_x, offset_y)
        
class ScrollView(FrameLayout):
    '''滚动视图
    '''

    @property
    def reach_top(self):
        '''滑动区域达到顶部
        '''
        scroll_rect = self._driver.get_control_scroll_rect(self.hashcode)
        return scroll_rect[1] == 0

    @property
    def reach_bottom(self):
        '''滑动区域达到底部
        '''
        scroll_rect = self._driver.get_control_scroll_rect(self.hashcode)
        return scroll_rect[1] + scroll_rect[3] >= scroll_rect[5]
    
    def _scroll(self, x, y, count=5, interval=0.04):
        '''横向或纵向滚动
        
        :param x: x > 0 时向左滑动，x = x1 - x2，滚动条向右
        :type x:  int
        :param y: y > 0 时向上滑动，y = y1 - y2，滚动条向下
        :type y:  int
        :param    count: 分为几次滑动
        :type     count:  int
        '''
        if y != 0: y = y * 100 // abs(y) if abs(y) < 100 else y  # 为防止在某个控件内滚动变成点击,设置最小滑动距离为100
        rect = self.rect
        
        mid_x = rect[0] + rect[2] // 2  # 中点
        mid_y = rect[1] + rect[3] // 2
        
        interval *= 1000  # 秒转换成毫秒
        if x != 0:
            x1 = mid_x + x // 2
            x2 = mid_x - x // 2
        else:
            x1 = x2 = mid_x
        
        if y != 0:
            y1 = mid_y + y // 2
            y2 = mid_y - y // 2
        else:
            y1 = y2 = mid_y
            
        self._driver.drag(x1, y1, x2, y2, count, interval)
        
    def scroll(self, x, y, count=5, interval=0.04):
        '''横向或纵向滚动
        
        :param x: x > 0 时向左滑动，x = x1 - x2，滚动条向右
        :type x:  int
        :param y: y > 0时向上滑动，y = y1 - y2，滚动条向下
        :type y:  int
        :param    count: 分为几次滑动
        :type     count:  int
        '''
        # 为避免在不可滑动区域滑动，每次滑动只在可滑动区域3/4处滑动
        time0 = time.time()
        while time.time() - time0 < 10:
            rect = self.rect
            if rect[2] == 0 or rect[3] == 0:
                time.sleep(0.5)
            else:
                break
            
        max_x = rect[2] * 3 // 4
        max_y = rect[3] * 3 // 4
        
        while abs(x) >= max_x:
            self._scroll(max_x * x // abs(x), 0)
            x -= max_x * x // abs(x)
        if x != 0:self._scroll(x, 0)
        
        while abs(y) >= max_y:
            self._scroll(0, max_y * y // abs(y))
            y -= max_y * y // abs(y)
        if y != 0:self._scroll(0, y)
        
    def on_scroll(self, x, y):
        time.sleep(0.5)

    def scroll_up_one_page(self):
        '''向上滑动一页
        '''
        if self.reach_top: return False
        rect = self.rect
        scroll_y = rect[3]
#        scroll_rect = self.driver.get_control_scroll_rect(self.hashcode)
#        if scroll_y > scroll_rect[1]:
#            scroll_y = scroll_rect[1]
        self.scroll(0, -scroll_y)
        self.on_scroll(0, -scroll_y)
        return True

    def scroll_down_one_page(self):
        '''向下滑动一页
        '''
        if self.reach_bottom: return False
        rect = self.rect
        scroll_y = rect[3]
#        scroll_rect = self.driver.get_control_scroll_rect(self.hashcode)
#        offset = scroll_rect[5] - scroll_rect[1] + scroll_rect[3]
#        if scroll_y > offset:
#            scroll_y = offset
        self.scroll(0, scroll_y)
        self.on_scroll(0, scroll_y)
        return True

    def scroll_to_top(self):
        '''滑动到顶部
        '''
        if self.reach_top: return
        rect = self.rect
        while not self.reach_top:
            self.scroll(0, -rect[3])
            self.on_scroll(0, -rect[3])
        self._wait_for_refresh_complete()  # 可能会出现刷新操作

    def scroll_to_bottom(self):
        '''滑动到底部
        '''
        if self.reach_bottom: return
        rect = self.rect
        while not self.reach_bottom:
            self.scroll(0, rect[3])
            self.on_scroll(0, rect[3])

    def _wait_for_refresh_complete(self, timeout=20):
        '''等待刷新完成
        '''
        outer_top = self.rect[1]
        if len(self.children) > 0:
            time0 = time.time()
            while time.time() - time0 < timeout:
                children = self.children
                if len(children) > 0:  # 没有子节点时等待
                    try:
                        inner_top = children[0].rect[1]
                    except RuntimeError as e:
                        logger.warn('_wait_for_refresh_complete error: %s' % e)
                        time.sleep(0.5)
                        continue
                    if inner_top <= outer_top: return True
                time.sleep(0.5)
        return False

    def pull_down_refresh(self):
        '''下拉刷新
        '''
        rect = self.rect
        while rect[2] == 0 or rect[3] == 0:
            time.sleep(0.1)
            rect = self.rect
        self.scroll_to_top()
        # self.scroll(0, -rect[3], 10, 0.1)  # 通过增加滑动次数增加时间，一般下拉时需要暂停一下才会触发刷新操作
        x = rect[0] + rect[2] // 2
        y1 = rect[1] + rect[3] // 4
        y2 = rect[1] + rect[3] * 3 // 4
        
        self._driver.drag(x, y1, x, y2, send_down_event=True, send_up_event=False)
        time.sleep(1)  # 按住不动
        self._driver.drag(x, y2, x, y2, send_down_event=False, send_up_event=True)
        
        self._wait_for_refresh_complete()
        time.sleep(1.5)  # 有些控件刷新完立即操作会不成功
        
class AbsListView(ScrollView):
    '''ListView和GridView基类
    '''
    def __init__(self, *args, **kwds):
        super(AbsListView, self).__init__(*args, **kwds)
        self._first_visible_position = 0  # 可见的第一个控件索引
        self._last_visible_position = 0  # 可见的最后一个索引
        self._item_count = 0  # 所有子节点个数
        self._children = []
        self._first_update = True  # 第一次更新时需要多做些操作

    def __iter__(self):
        '''迭代器
        '''
        self.update()
        for i in range(self.item_count):
            if i < self.first_position:
                # 需要往上滑动
                while i < self.first_position:
                    self.scroll_up_one_page()
                if i > self.last_position:  # 防止滚过
                    self.scroll(0, 100 * (i - self.last_position + 1))
                    self.update()
            elif i > self.last_position:
                # 需要往下滑动
                while i > self.last_position:
                    if i >= self.item_count: return
                    self.scroll_down_one_page()
                if i < self.first_position:
                    self.scroll(0, -100 * (self.first_position - i + 1))
                    self.update()

            idx = i - self.first_position
            if idx >= len(self._children):
                # 可能之前拉到的数据不正确
                self.update()
            if idx < 0 or idx >= len(self._children):
                raise IndexError('%d 不在范围[0, %d]中' % (idx, len(self._children) - 1))
            yield ListItem(self._children[idx])

    def __len__(self):
        self.update()
        return self.item_count

    def __getitem__(self, key):
        '''支持listview[i]方式访问子控件
        '''
        if isinstance(key, int):
            return self.get_child(key)
        else:
            raise TypeError('只支持整型索引')

    def __str__(self):
        return '<%s(Count=%d) at 0x%X>' % (self.__class__.__name__, len(self), id(self))

    @property
    def item_count(self):
        return self._item_count

    @property
    def first_position(self):
        return self._first_visible_position

    @property
    def last_position(self):
        return self._last_visible_position

    def on_scroll(self, x, y):
        super(AbsListView, self).on_scroll(x, y)
        self.update()
        if self.reach_top:
            self._wait_for_refresh_complete()

    @func_wrap
    def _update_list(self):
        listview_info = self._driver.get_listview_info(self.hashcode)
        self._item_count = listview_info['Count']
        self._first_visible_position = listview_info['FirstPosition']
        self._last_visible_position = listview_info['LastPosition']
        if self._item_count > 0:
            self._children = self.children  # 防止每次都访问

    def update(self):
        for _ in range(20):
            show_count = self._last_visible_position - self._first_visible_position + 1
            self._update_list()
            if self._first_visible_position >= 0 and self._last_visible_position >= 0 and len(self._children) == self._last_visible_position - self._first_visible_position + 1:
                # 防止控件树还在构建过程
                if not self._first_update:
                    show_count_now = self._last_visible_position - self._first_visible_position + 1
                    if show_count_now == show_count: return  # 刷新前后显示的节点数目不变才是稳定状态
                self.wait_for_complete()
                self._first_update = False
                return
            time.sleep(0.1)

    def get_child(self, idx):
        '''提供使用索引访问子元素的方法
        '''
        self.update()
        item_count = self.item_count
        if idx < 0:
            # 允许使用负数索引
            idx += item_count
        if idx < 0 or idx >= item_count:
            raise IndexError('index=%d超出list范围' % idx)
        if idx < self.first_position:
            # 需要往上滑动
            while idx < self.first_position:
                self.scroll_up_one_page()
        elif idx > self.last_position:
            # 需要往下滑动
            while idx > self.last_position:
                self.scroll_down_one_page()
        idx = idx - self.first_position
        if idx < 0 or idx >= len(self._children):
            raise IndexError('index=%d错误，不在范围[0, %d]中' % (idx, len(self._children) - 1))
        return ListItem(self._children[idx], idx)
    
    @property
    def reach_top(self):
        return self.reached_top
    
    @property
    def reached_top(self):
        '''滑动区域达到顶部
        '''
        for _ in range(3):
            self.update()
            if self.first_position > 0: return False
            if self.item_count == 0: return True
            if len(self._children) == 0: return True
            # 此时第一个子节点肯定可见
            rect = View.rect.fget(self)  # 子类可能会重载rect以修改滑动范围，此时不能根据子类的实现来计算
            try:
                crect = self._children[0].rect
            except RuntimeError as e:
                from qt4a.androiddriver.util import logger
                # logger.warn('reach_top error: %s' % e)
                logger.exception(e)
                continue
            return crect[1] >= rect[1]
    
    @property
    def reach_bottom(self):
        return self.reached_bottom
    
    @property
    def reached_bottom(self):
        '''滑动区域达到底部
        '''
        if self.item_count == 0: self.update()
        return self.last_position == self.item_count - 1

#    def scroll_to_top(self):
#        '''滑动到顶部
#        '''
#        if self.reach_top: return
#        rect = self.bounding_rect
#        scroll_total = 0
#        while not self.reach_top:
#            scroll_y = self.first_position * 100  # 以每一项100的高度计算
#            if scroll_y > rect[3]: scroll_y = rect[3]
#            print ('scroll', -scroll_y)
#            self.scroll(0, -scroll_y)
#            scroll_total += scroll_y
#        self.on_scroll(0, -scroll_total)
#    
    def scroll_up_one_page(self):
        '''向上滑动一页
        '''
        if self.reached_top: return False
        rect = self.rect
        scroll_y = rect[3]
        if scroll_y > self.first_position * 100:
            scroll_y = self.first_position * 100
        if scroll_y < 100: scroll_y = 100
        self.scroll(0, -scroll_y)
        self.on_scroll(0, -scroll_y)
        return True

    def scroll_down_one_page(self):
        '''向下滑动一页
        '''
        if self.reached_bottom: return False
        rect = self.rect
        scroll_y = rect[3]
        offset = (self.item_count - self.last_position) * 100
        if scroll_y > offset:
            scroll_y = offset
        self.scroll(0, scroll_y)
        self.on_scroll(0, scroll_y)
        return True

    def wait_for_complete(self, timeout=2):
        '''等待ListView控件变化，比如需要读取本地数据
        '''
        count0 = self.item_count
        first_position0 = self.first_position
        last_position0 = self.last_position

        time0 = time.time()
        while time.time() - time0 < timeout:
            time.sleep(0.2)
            self._update_list()
            if self._first_visible_position >= 0 and self._last_visible_position >= 0 and len(self._children) == self._last_visible_position - self._first_visible_position + 1:
                count = self.item_count
                first_position = self.first_position
                last_position = self.last_position
                if count == count0 and first_position == first_position0 and last_position == last_position0: return True
                count0 = count
                first_position0 = first_position
                last_position0 = last_position
        return False

class ListItem(View):
    '''为方便遍历AbsListView，表示AbsListView的直接子孩子
    '''
    def __init__(self, view, idx=None):
        '''
        :param view: AbsListView中每一个子节点
        :type view:  View
        '''
        self._view = view
        self._control_dict = {}  # 用于缓存
        self._idx = idx  # 项索引，用于重新查找控件
        
    def __getitem__(self, key):
        '''支持listitem[key]方式访问子节点
        :param key: 封装时定义的子节点名
        :type key:  string
        '''
        if key in self._control_dict:
            return self._control_dict[key]
        item = self._view.container[key]
        import copy
        result = copy.copy(item)  # 复制一个对象
        result._root = self  # 修改根节点为当前节点
        self._control_dict[key] = result
        return result

    def __getattr__(self, attr):
        '''委托self._view实现
        '''
        return getattr(self._view, attr)
    
    def _update(self):
        '''如果没有索引值直接抛异常
        '''
        from qt4a.androiddriver.util import logger
        logger.debug('ListItem update, idx=%s' % self._idx)
        if self._idx == None:
            raise RuntimeError('ListView无法重新获取控件')
        self._view._root.update()
        self._view = self._view._root._children[self._idx]  # 重新获取Item项
    
    def has(self, key):
        '''是否存在该节点
        
        :param key: 封装时定义的子节点名
        :type  key: string
        '''
        from qt4a.androiddriver.util import ControlExpiredError, logger
        item = self._view.container[key]
        # child = self._driver.get_control(self._activity, self.hashcode, item._locator)
        self._view._need_convert_qpath = True  # 避免判断多个节点存在时不会使用整型ID查找问题
        try:
            hashcode = self._view._get_hashcode(self._view.hashcode, item._locator)
            if hashcode == 0: return False
        except ControlExpiredError:
            logger.info('ControlExpiredError occur in has')
            if self._idx == None: return False
            self._update()
            return self.has(key)
        item._locator = self._locator  # 替换为真正的QPath
        return True

class ListView(AbsListView):
    '''列表视图
    '''
    pass

class GridView(AbsListView):
    '''格子视图
    '''

class TabWidget(View):
    '''Tab控件
    '''
    @property
    def selected_index(self):
        '''当前所选项的索引
        '''
        return self._driver.get_selected_tab_index(self.hashcode)

class WebkitWebView(View):
    '''4.4以下版本中的WebView
    '''
    
    def __init__(self, *args, **kwargs):
        super(WebkitWebView, self).__init__(*args, **kwargs)
    
    def eval_script(self, frame_xpaths, script):
        '''执行JavaScript
        '''
        from qt4a.androiddriver.util import AndroidSpyError, ControlExpiredError
        try:
            return self._driver.eval_script(self.hashcode, frame_xpaths, script)
        except ControlExpiredError as e:
            raise e
        except AndroidSpyError as e:
            try:
                from qt4w.util import JavaScriptError
            except ImportError:
                raise
            else:
                raise JavaScriptError(frame_xpaths, e.args[0])
            
class WebView(View):
    '''Web页面容器
    '''
    title = None
    url = None
    is_chromium = None
    
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        super(WebView, self).__init__(*args, **kwargs)
        self._is_chromium = self.__class__.is_chromium
        if self._is_chromium == None: self._is_chromium = self.container.device.sdk_version >= 19  # Android 4.4以上使用Chromium内核
        self._webview_impl = None
    
    @property
    def webdriver_class(self):
        '''WebView对应的WebDriver类
        '''
        from qt4a.androiddriver.webdriver import AndroidWebDriver
        return AndroidWebDriver
    
    @property
    def webview_impl(self):
        '''
        '''
        if self._webview_impl == None:
            if not self._is_chromium:
                self._webview_impl = WebkitWebView(*self._args, **self._kwargs)
            else:
                try:
                    self._webview_impl = ChromiumWebView(title=self.__class__.title, url=self.__class__.url, *self._args, **self._kwargs)
                except ImportError:
                    # Fallback to custom WebView
                    self._webview_impl = WebkitWebView(*self._args, **self._kwargs)

        return self._webview_impl
    
    def eval_script(self, frame_xpaths, script):
        '''在指定frame中执行JavaScript，并返回执行结果（该实现需要处理js基础库未注入情况的处理）
        
        :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，怎传入“[]”
        :type frame_xpaths:  list
        :param script:       要执行的JavaScript语句
        :type script:        string
        '''
        return self.webview_impl.eval_script(frame_xpaths, script)
    
    @property
    def visible_rect(self):
        '''WebView控件可见区域的坐标信息
        '''
        return self.rect
            
    def disable_soft_input(self):
        '''禁用软键盘
        '''
        return self._driver.enable_soft_input(self.hashcode, False)
    
    def click(self, x_offset=None, y_offset=None):
        '''点击WebView中的某个坐标
        
        :param x_offset: 与WebView左上角的横向偏移量
        :type x_offset:  int/float
        :param y_offset: 与WebView左上角的纵向偏移量
        :type y_offset:  int/float
        '''
        if x_offset != None or y_offset != None:
            rect = self.rect
            x = (rect[0] + rect[2] // 2) if not x_offset else (rect[0] + x_offset)
            y = (rect[1] + rect[3] // 2) if not y_offset else (rect[1] + y_offset)
            return self._click(0, x, y)
        else:
            return super(WebView, self).click()
    
    def long_click(self, x_offset, y_offset, duration=1):
        '''长按WebView中的某个坐标
        
        :param x_offset: 与WebView左上角的横向偏移量
        :type x_offset:  int/float
        :param y_offset: 与WebView左上角的纵向偏移量
        :type y_offset:  int/float
        :param duration: 按住的持续时间
        :type duration:  int/float
        '''
        return super(WebView, self).long_click(duration, x_offset, y_offset)
    
    def send_keys(self, keys):
        '''向WebView中的输入框发送按键
        
        :param keys: 要发送的按键
        :type keys:  string
        '''
        self._driver.send_keys(keys)
    
    def drag(self, x1, y1, x2, y2):
        '''从(x1, y1)点滑动到(x2, y2)点
        
        :param x1: 起点横坐标
        :type x1:  int/float
        :param y1: 起点纵坐标
        :type y1:  int/float
        :param x2: 终点横坐标
        :type x2:  int/float
        :param y2: 终点纵坐标
        :type y2:  int/float
        '''
        self._driver.drag(x1, y1, x2, y2, 10, 0.1)
    
    def screenshot(self):
        '''当前WebView的截图
        :return: PIL.Image
        '''
        from PIL import Image
        temp_path = tempfile.mktemp('.jpg')
        self.container.device.take_screen_shot(temp_path)
        with open(temp_path, 'rb') as fp:
            image_data = fp.read()
        os.remove(temp_path)
        image = Image.open(BytesIO(image_data))
        image = image.crop(self.rect)
        return image
    
class ChromiumWebView(WebkitWebView):
    '''Chromium内核的WebView
    '''
    debugger_instances = {}
    
    def __init__(self, url=None, title=None, *args, **kwargs):
        import chrome_master
        super(ChromiumWebView, self).__init__(*args, **kwargs)
        self._is_system_webview = False
        webview_types = self._driver.get_control_type(self.hashcode)
        for webview_type in webview_types:
            if webview_type == 'android.webkit.WebView':
                self._is_system_webview = True
                break
        self._url = url
        self._title = title
        
        self._device = self.container.device
        self._pid = self._device.adb.get_pid(self._driver._process_name)
        self._service_name = 'webview_devtools_remote_%d' % self._pid
        chrome_master.set_logger(logger)
        self._chrome_master = chrome_master.ChromeMaster(('localhost', 80), self.create_socket) #高版本Chromium内核必须要使用localhost的Host

        if self._pid in self.__class__.debugger_instances and self.__class__.debugger_instances[self._pid]:
            self.__class__.debugger_instances[self._pid].close()
            self.__class__.debugger_instances[self._pid] = None

        self._page_debugger = self.get_debugger()
        self._page_debugger.register_handler(chrome_master.RuntimeHandler)
        self.__class__.debugger_instances[self._pid] = self._page_debugger

    def create_socket(self):
        '''创建socket对象
        '''
        try:
            sock = self._device.adb.create_tunnel(self._service_name, 'localabstract')
            if sock: return sock
        except:
            logger.warn('create socket tunnel %s failed' % self._service_name)
        self._driver.set_webview_debugging_enabled(self.hashcode, True)
        return self.create_socket()
        
    def get_page_url(self):
        '''获取当前WebView页面的url
        '''
        if not self._is_system_webview: return self._url
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            try:
                ret = super(ChromiumWebView, self).eval_script([], 'location.href')
            except:
                logger.exception('Read page url failed')
                return
            if ret != 'about:blank': return ret
            time.sleep(0.5)
        else:
            raise RuntimeError('Get page url failed')
    
    def get_debugger(self):
        '''获取当前页面的RemoteDebugger实例
        '''
        url = self.get_page_url()
        return self._chrome_master.find_page(url=url, title=self._title)  # 获取最后一个页面
    
    def convert_frame_tree(self, frame_tree, parent=None):
        '''将frame tree转化为Frame对象
        '''
        from qt4w import util
        frame = util.Frame(frame_tree['frame']['id'], frame_tree['frame'].get('name'), frame_tree['frame']['url'])
        if parent: parent.add_child(frame)
        if 'childFrames' in frame_tree:
            for child in frame_tree['childFrames']:
                self.convert_frame_tree(child, frame)
        return frame
        
    def _get_frame_id_by_xpath(self, frame_xpaths, timeout=10):
        '''获取frame id
        '''
        from qt4w import util
        time0 = time.time()
        while time.time() - time0 < timeout:
            frame_tree = self._page_debugger.page.get_frame_tree()
            frame = self.convert_frame_tree(frame_tree)
            frame_selector = util.FrameSelector(self, frame)
            try:
                frame = frame_selector.get_frame_by_xpath(frame_xpaths)
            except util.ControlNotFoundError:
                pass
            else:
                return frame.id
            time.sleep(0.5)
        else:
            raise ControlNotFoundError('Find frame %s timeout' % ''.join(frame_xpaths))
        
    def eval_script(self, frame_xpaths, script):
        '''在指定frame中执行JavaScript，并返回执行结果
        
        :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，怎传入“[]”
        :type frame_xpaths:  list
        :param script:       要执行的JavaScript语句
        :type script:        string
        '''
        from chrome_master import util
        frame_id = self._get_frame_id_by_xpath(frame_xpaths)
        try:
            return self._page_debugger.runtime.eval_script(frame_id, script)
        except util.JavaScriptError as e:
            from qt4w.util import JavaScriptError
            raise JavaScriptError(frame_xpaths, e.message)
        
class ViewPager(ViewGroup):
    '''横向滚动视图
    '''
    @property
    def current_item_index(self):
        '''当前显示项的索引
        '''
        ret = self._driver.get_object_field_value(self.hashcode, 'mCurItem')
        return int(ret)
    
    @property
    def item_count(self):
        '''总项数
        '''
        ret = self._driver.get_object_field_value(self.hashcode, 'mItems.size')
        return int(ret)
    
    def scroll(self, count=1, rect=None):
        '''左右滚动
        
        :param count: 滑动次数，大于0表示向右滑动，小于0表示向左滑动
        :type  count: int
        '''
        if rect == None: rect = self.rect
        x1 = rect[0] + 5
        x2 = rect[0] + rect[2] - 5
        if count > 0:
            # 交换终点
            x = x1
            x1 = x2
            x2 = x
        y = rect[1] + rect[3] // 2
        for _ in range(abs(count)):
            self._driver.drag(x1, y, x2, y, 1)
    
    def __iter__(self):
        '''迭代器
        '''
        for i in range(self.item_count):
            idx = self.current_item_index
            if idx > i:
                # 向右滑动
                self.scroll(i - idx)
            elif idx < i:
                # 向左滑动
                self.scroll(idx - i)
            yield ListItem(self.children[i])
            
class RadioGroup(LinearLayout):
    '''
    '''
    pass

class RecyclerView(AbsListView):
    '''虽然不是继承自ListView，但是可以复用其部分代码，所以从ListView继承过来
    '''
    @func_wrap
    def _update_list(self):
        self._item_count = self._driver.call_object_method(self.hashcode, 'mAdapter', 'getItemCount')
        self._first_visible_position = -1
        self._last_visible_position = -1
        if self._item_count > 0:
            self._children = self.children  # 防止每次都访问
        if len(self._children) > 0:
            # 获取第一个子节点的索引值
            self._first_visible_position = self._get_first_visible_position()
            self._last_visible_position = self._first_visible_position + len(self._children) - 1
    
    def _get_first_visible_position(self):
        '''获取第一个可见节点的位置
        ('mLayoutParams', 'getViewLayoutPosition')
        '''
        if hasattr(self, '_get_first_visible_position_args'):
            return self._driver.call_object_method(self._children[0].hashcode, *self._get_first_visible_position_args)
        
        for args in (('mLayoutParams.mViewHolder', 'getPosition'),
                     ('mLayoutParams.mViewHolder', 'getLayoutPosition')):
            try:
                result = self._driver.call_object_method(self._children[0].hashcode, *args)
                self._get_first_visible_position_args = args
                return result
            except:
                logger.exception('get first visible position by %r failed' % args)
        
    def _is_horizontal(self):
        '''是否横向滑动
        '''
        if not hasattr(self, '__horizontal'):
            if self.item_count >= 2:
                self.__horizontal = self.children[0].rect[0] != self.children[1].rect[0]
            elif self.item_count == 1:
                x_ratio = 1.0 * self.children[0].rect[2] / self.rect[2]
                y_ratio = 1.0 * self.children[0].rect[3] / self.rect[3]
                self.__horizontal = x_ratio < y_ratio
            else:
                return False
        return self.__horizontal
    
    @property
    def reached_left(self):
        '''滑动区域达到最左边
        '''
        for _ in range(3):
            self.update()
            if self.first_position > 0: return False
            if self.item_count == 0: return True
            if len(self._children) == 0: return True
            # 此时第一个子节点肯定可见
            rect = View.rect.fget(self)  # 子类可能会重载rect以修改滑动范围，此时不能根据子类的实现来计算
            try:
                crect = self._children[0].rect
            except RuntimeError as e:
                logger.exception(e)
                continue
            return crect[0] >= rect[0]

    @property
    def reached_right(self):
        '''滑动区域达到最右边
        '''
        if self.item_count == 0: self.update()
        return self.last_position == self.item_count - 1
        
    def scroll_left_one_page(self):
        '''向左滑动一页
        '''
        if self.reached_left: return False
        rect = self.rect
        scroll_x = rect[2]
        item_rect = self._children[0].rect
        if scroll_x > self.first_position * item_rect[2]:
            scroll_x = self.first_position * item_rect[2]
        if scroll_x < 100: scroll_x = 100
        self.scroll(-scroll_x, 0)
        self.on_scroll(-scroll_x, 0)
        return True

    def scroll_right_one_page(self):
        '''向右滑动一页
        '''
        if self.reached_right: return False
        rect = self.rect
        scroll_x = rect[2]
        item_rect = self._children[0].rect
        offset = (self.item_count - self.last_position) * item_rect[2]
        if scroll_x > offset:
            scroll_x = offset
        self.scroll(scroll_x, 0)
        self.on_scroll(scroll_x, 0)
        return True
    
    def scroll_up_one_page(self):
        '''向上滑动一页
        '''
        if self.reached_top: return False
        rect = self.rect
        scroll_y = rect[3]
        child_rect = self._children[0].rect
        if scroll_y > self.first_position * child_rect[3]:
            scroll_y = self.first_position * child_rect[3]
        if scroll_y < child_rect[3]: scroll_y = child_rect[3]
        self.scroll(0, -scroll_y)
        self.on_scroll(0, -scroll_y)
        return True

    def scroll_down_one_page(self):
        '''向下滑动一页
        '''
        if self.reached_bottom: return False
        rect = self.rect
        scroll_y = rect[3]
        child_rect = self._children[0].rect
        offset = (self.item_count - self.last_position) * child_rect[3]
        if scroll_y > offset:
            scroll_y = offset
        elif scroll_y < child_rect[3]:
            scroll_y = child_rect[3]
        self.scroll(0, scroll_y)
        self.on_scroll(0, scroll_y)
        return True
    
    def _scroll_child_to_visible(self, child_idx):
        '''将子节点滚动到可见区域
        '''
        if child_idx < self.first_position:
            # 需要往左/上滑动
            while child_idx < self.first_position:
                if self._is_horizontal():
                    self.scroll_left_one_page()
                else:
                    self.scroll_up_one_page()
                self._update_list()
        elif child_idx > self.last_position:
            # 需要往右/下滑动
            while child_idx > self.last_position:
                if self._is_horizontal():
                    self.scroll_right_one_page()
                else:
                    self.scroll_down_one_page()
                self._update_list()
                
        child = self.children[child_idx - self.first_position]
        self_rect = self.rect
        child_rect = child.rect
        
        # 保证当前项完全可见
        
        if child_rect[0] < self_rect[0]: 
            # 向左滑动
            scroll_x = self_rect[0] - child_rect[0]
            self.scroll(-scroll_x, 0)
        elif child_rect[0] + child_rect[2] > self_rect[0] + self_rect[2]: 
            # 向右滑动
            scroll_x = child_rect[0] + child_rect[2] - (self_rect[0] + self_rect[2])
            self.scroll(scroll_x, 0)
        elif child_rect[1] < self_rect[1]: 
            # 向上滑动
            scroll_y = self_rect[1] - child_rect[1]
            self.scroll(0, -scroll_y)
        elif child_rect[1] + child_rect[3] > self_rect[1] + self_rect[3]:
            # 向下滑动
            scroll_y = child_rect[1] + child_rect[3] - (self_rect[1] + self_rect[3])
            self.scroll(0, scroll_y)
        else:
            return
        self._update_list()
        
    def get_child(self, idx):
        '''提供使用索引访问子元素的方法
        '''
        self.update()
        item_count = self.item_count
        if idx < 0:
            # 允许使用负数索引
            idx += item_count
        if idx < 0 or idx >= item_count:
            raise IndexError('index=%d超出list范围' % idx)
        self._scroll_child_to_visible(idx)
        idx = idx - self.first_position
        if idx < 0 or idx >= len(self._children):
            raise IndexError('index=%d错误，不在范围[0, %d]中' % (idx, len(self._children) - 1))
        return ListItem(self._children[idx], idx)
    
    def __iter__(self):
        '''迭代器
        '''
        self.update()
        for i in range(self.item_count):
            self._scroll_child_to_visible(i)

            idx = i - self.first_position
            if idx >= len(self._children):
                # 可能之前拉到的数据不正确
                self.update()
            if idx < 0 or idx >= len(self._children):
                raise IndexError('%d 不在范围[0, %d]中' % (idx, len(self._children) - 1))
            yield ListItem(self._children[idx])

class DatePicker(FrameLayout):
    '''时间选择器
    '''
    def set_date(self, year, month, day): 
        if year < 0 or month < 1 or month > 12 or day < 1 or day > 31:
            raise RuntimeError('参数传入不符合时间要求')
#         if self._driver._device.sdk_version < 21:
        if self._driver._device.adb.get_sdk_version() < 21:  # 5.0以下用如下方法
            self._driver.call_object_method(self.hashcode, 'mCurrentDate', 'set', 'void', year, month - 1, day)
        else:
            self._driver.call_object_method(self.hashcode, 'mDelegate.mCurrentDate', 'set', 'void', year, month - 1, day)

class ActionMenuItemView(TextView):
    '''android.support.v7.internal.view.menu.ActionMenuItemView
    '''
    pass

class AppCompatEditText(EditText):
    '''android.support.v7.widget.AppCompatEditText
    '''
    pass

class AppCompatImageView(ImageView):
    '''android.support.v7.widget.AppCompatImageView
    '''
    pass

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''AI控件库
'''

from io import BytesIO
import os
import tempfile
from qt4a.andrcontrols import Window, View

class MtWindow(Window):
    # TODO: to be deleted
    pass
    
try:
    from metislib.controls import *
except ImportError as e:
    class MtControl(View):
        pass
else:
    class MetisView(object):
        '''实现IMetisView接口
        '''
        def __init__(self, view_or_window):
            if not isinstance(view_or_window, Window) and not isinstance(view_or_window, View):
                raise TypeError('view_or_window must be Window or View object')
            self._view_or_window = view_or_window
            if isinstance(self._view_or_window, View):
                self._device = self._view_or_window.container.device
            else:
                self._device = self._view_or_window.device
                
        @property
        def rect(self):
            '''元素相对坐标(x, y, w, h)
            '''
            if isinstance(self._view_or_window, View):
                return self._view_or_window.rect
            else:
                screen_size = self._device.screen_size
                return (0, 0, screen_size[0], screen_size[1])
            
        @property
        def os_type(self):
            '''系统类型，例如"android"，"ios"，"pc"
            '''
            return 'android'
    
        def screenshot(self):
            '''当前容器的区域截图
            :return: PIL.image
            '''
            from PIL import Image
            clip = False
            if isinstance(self._view_or_window, View): clip = True
            temp_path = tempfile.mktemp('.jpg')
            self._device.take_screen_shot(temp_path)
            with open(temp_path, 'rb') as fp:
                image_data = fp.read()
            os.remove(temp_path)
            image = Image.open(BytesIO(image_data))
            if clip:
                x, y, w, h = self.rect
                image = image.crop((x, y, x + w, y + h))
            return image
        
        def _get_position(self, offset_x=None, offset_y=None):
            '''
            '''
            if offset_x == None: offset_x = 0.5
            if offset_y == None: offset_y = 0.5
            rect = self.rect
            x = rect[0] + int(rect[2] * offset_x)
            y = rect[1] + int(rect[3] * offset_y)
            return x, y
        
        def click(self, offset_x=None, offset_y=None):
            '''点击
            :param offset_x: 相对于该控件的坐标offset_x，百分比( 0 -> 1 )，不传入则默认该控件的中央
            :type offset_x: float|None
            :param offset_y: 相对于该控件的坐标offset_y，百分比( 0 -> 1 )，不传入则默认该控件的中央
            :type offset_y: float|None
            '''
            x, y = self._get_position(offset_x, offset_y)
            self._device.run_shell_cmd('input tap %d %d' % (x, y))

        def send_keys(self, text):
            '''
            '''
            self._device.send_text(text)
    
        def double_click(self, offset_x=None, offset_y=None):
            '''
            '''
            x, y = self._get_position(offset_x, offset_y)
            self._device.run_shell_cmd('input tap %s %s' % (x, y))
            self._device.run_shell_cmd('input tap %s %s' % (x, y))
    
        def long_click(self, duration, offset_x=None, offset_y=None):
            '''
            '''
            x, y = self._get_position(offset_x, offset_y)
            self._device.run_shell_cmd('input swipe %s %s %s %s %d' % (x, y, x, y, int(duration * 1000)))
            
        def drag(self, from_x=0.5, from_y=0.5, to_x=0.5, to_y=0.1, duration=0.5):
            '''拖拽
            
            :param from_x: 起点 x偏移百分比（从左至右为0.0至1.0）
            :type from_x: float
            :param from_y: 起点 y偏移百分比（从上至下为0.0至1.0）
            :type from_y: float
            :param to_x: 终点 x偏移百分比（从左至右为0.0至1.0）
            :type to_x: float
            :param to_y: 终点 y偏移百分比（从上至下为0.0至1.0）
            :type to_y: float
            :param duration: 持续时间（秒）
            :type duration: float
            '''
            from_x, from_y = self._get_position(from_x, from_y)
            to_x, to_y = self._get_position(to_x, to_y)
            return self._device.drag(from_x, from_y, to_x, to_y, count=5, wait_time=100, send_down_event=True, send_up_event=True)
    
        
if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''ADB客户端，用于与ADB守护进程通信
'''

from __future__ import unicode_literals

import six
import os
import time
import socket, select
import struct
import threading
from io import BytesIO
from qt4a.androiddriver.util import logger, TimeoutError

SYNC_DATA_MAX = 64 * 1024

class AdbError(RuntimeError):
    pass

class Pipe(object):
    '''模拟实现内存管道
    '''
    def __init__(self):
        self._buffer = BytesIO()
        self._max_buffer_size = 4096 * 16
        self._lock = threading.Lock()
        self._pos = 0  # 当前读指针位置
        self._write_buffer = b''  # 保证每次都是整行写
        
    def write(self, s):
        self._write_buffer += s
        pos = self._write_buffer.rfind(b'\n')
        if pos <= 0: return
        s = self._write_buffer[:pos]
        self._write_buffer = self._write_buffer[pos:]
        with self._lock:
            self._buffer.seek(0, 2)  # 将指针置于尾部
            self._buffer.write(s)
    
    def readline(self):
        wait = False
        while True:
            if wait: time.sleep(0.1)
            with self._lock:
                self._buffer.seek(0, 2)
                buffer_size = self._buffer.tell()
                if buffer_size <= self._pos:
                    wait = True
                    continue
            with self._lock:
                self._buffer.seek(self._pos)
                ret = self._buffer.readline()
                if len(ret) == 0:
                    wait = True
                    continue
                else:
                    self._pos = self._buffer.tell()
                    self._buffer.seek(0, 2)
                    buffer_size = self._buffer.tell()
                    if buffer_size >= self._max_buffer_size:
                        # 创建新的缓冲区
                        self._buffer.seek(self._pos)
                        buffer = self._buffer.read()
                        self._buffer.close()
                        self._buffer = BytesIO()
                        self._buffer.write(buffer)
                        self._pos = 0
                    return ret
    
    def read(self):
        '''读取管道中的所有数据
        '''
        with self._lock:
            self._buffer.seek(self._pos)
            result = self._buffer.read()
            if self._write_buffer: 
                result += self._write_buffer
                self._write_buffer = ''
            return result
        
class ADBPopen(object):
    '''与Popen兼容
    '''
    class StdinPipe(object):
        '''
        '''
        def __init__(self, sock):
            self._sock = sock
            
        def write(self, s):
            self._sock.send(s)
        
        def flush(self):
            pass

    def __init__(self, sock, timeout=None):
        self._sock = sock
        self._stdin = self.StdinPipe(sock)
        self._stdout = Pipe()
        self._stderr = Pipe()
        self._running = True
        self._timeout = timeout
        if self._timeout == None: self._timeout = 0xFFFFFFFF
        self._event = threading.Event()  # 接收完数据的事件通知
        self._thread = threading.Thread(target=self._work_thread, args=(), name=self.__class__.__name__)
        self._thread.setDaemon(True)
        self._thread.start()
        
    @property
    def stdin(self):
        return self._stdin
    
    @property
    def stdout(self):
        return self._stdout
    
    @property
    def stderr(self):
        return self._stderr
    
    @property
    def pid(self):
        return self._thread.ident
    
    def _work_thread(self):
        time0 = time.time()
        while self._running and time.time() - time0 < self._timeout:
            infds, outfds, errfds = select.select([self._sock, ], [], [], 1)
            if len(infds) > 0:
                try:
                    buff = self._sock.recv(4096)
                    if len(buff) == 0:
                        self._sock.close()
                        self._sock = None
                        self._running = False
                        self._event.set()
                        return
                    self._stdout.write(buff)
                except socket.error as e:
                    logger.info("接收返回数据错误： %s" % (e))
#                    import traceback
#                    traceback.print_exc()
                    self._stdout.write(b' ')  # 通知接收方退出
                    self._sock.close()
                    self._sock = None
                    self._running = False
                    self._event.set()
                    return
        self._sock.close()
        self._sock = None
        
    def poll(self):
        '''是否存在
        '''
        if self._thread.is_alive():
            return None
        else:
            return 0
    
    def terminate(self):
        '''结束
        '''
        self._running = False
        time.sleep(1)  # 等待线程退出
    
    def communicate(self):
        '''
        '''
        while True:
            if self._event.wait(0.001) == True or self.poll() == 0: 
                if self._running: raise TimeoutError('Execute timeout')
                return self.stdout.read(), self.stderr.read()
            # time.sleep(0.001)
            
class ADBClient(object):
    '''
    '''
    instance_dict = {}
    
    def __init__(self, server_addr='127.0.0.1', server_port=5037):
        self._server_addr = server_addr
        self._server_port = server_port
        self._sock = None
        self._lock = threading.Lock()

    @staticmethod
    def get_client(host, port=5037):
        '''根据主机名获取ADBClient实例
        '''
        return ADBClient(host, port)

    def call(self, cmd, *args, **kwds):
        '''调用命令字
        '''
        cmd = cmd.replace('-', '_')
        if cmd == 'forward' and args[1] == '--remove':
            method = getattr(self, 'remove_forward')
            args = list(args)
            args.pop(1)  # remove --remove args
        else:
            method = getattr(self, cmd)
        # print (args)
        sync = True
        if 'sync' in kwds: sync = kwds.pop('sync')
        if 'timeout' in kwds and not cmd in ('shell', 'install', 'uninstall', 'wait_for_device', 'reboot'): kwds.pop('timeout')
        if sync: 
            ret = None
            retry_count = kwds.pop('retry_count')
            i = 0
            socket_error_count = 0
            while i < retry_count:
                try:
                    self._lock.acquire()
                    ret = method(*args, **kwds)
                    break
                except socket.error as e:
                    logger.exception(u'执行%s %s error' % (cmd, ' '.join(args)))
                    socket_error_count += 1
                    if socket_error_count <= 10: i -= 1
                    time.sleep(1)
                except AdbError as e:
                    err_msg = str(e)
                    if 'device not found' in err_msg:
                        return '', 'error: device not found'
                    elif 'cannot bind to socket' in err_msg:
                        return '', err_msg
                    elif 'cannot remove listener' in err_msg:
                        return '', err_msg
                    elif 'device offline' in err_msg:
                        return '', 'error: device offline'
                    elif 'Bad response' in err_msg or 'Device or resource busy' in err_msg or 'closed' in err_msg:  # wetest设备有时候会返回closed错误
                        # 需要重试
                        logger.exception('Run %s%s %r' % (cmd, ' '.join(args), e))
                    else:
                        raise RuntimeError(u'执行%s %s 命令失败：%s' % (cmd, ' '.join(args), e))
                    time.sleep(1)
                    if i >= retry_count - 1: raise e
                except RuntimeError as e:
                    logger.exception(u'执行%s%s %r' % (cmd, ' '.join(args), e))
                    if 'device not found' in str(e):
                        self.wait_for_device(args[0], retry_count=1, timeout=300)
                        self._sock = None
                        return self.call(cmd, *args, **kwds)
                finally:
                    i += 1
                    if self._sock != None:
                        self._sock.close()
                        self._sock = None
                    self._lock.release()

            if ret == None: raise TimeoutError(u'Run cmd %s %s failed' % (cmd, ' '.join(args)))

            if isinstance(ret, (six.string_types, six.binary_type)):
                return ret, ''
            else:
                return ret
        else:
            self._transport(args[0])  # 异步操作的必然需要发送序列号
            if cmd == 'shell':
                self._lock.acquire()
                self._send_command('shell:' + ' '.join(args[1:]))
                pipe = ADBPopen(self._sock)
                self._sock = None
                self._lock.release()
                return pipe
            
        
    def _connect(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        for i in range(3):
            try:
                self._sock.connect((self._server_addr, self._server_port))
                return True
            except socket.error:
                pass
        return False

    def _check_status(self):
        '''检查返回状态
        '''
        stat = self._sock.recv(4)
        if stat == b"OKAY":
            return True
        elif stat == b"FAIL":
            size = int(self._sock.recv(4), 16)
            val = self._sock.recv(size)
            self._sock.close()
            self._sock = None
            raise AdbError(val.decode('utf8'))
        else:
            raise AdbError("Bad response: %r" % (stat,))

    def _send_command(self, cmd):
        data = "%04x%s" % (len(cmd), cmd)
        if not self._sock: self._connect()
        # logger.debug('send: %r' % data)
        self._sock.send(data.encode('utf8'))
        return self._check_status()

    def _recv(self, size=None):
        '''从socket读取数据
        '''
        result = b''
        if size != None:
            while len(result) < size:
                result += self._sock.recv(size - len(result))
        else:
            data = self._sock.recv(4096)
            while data:
                result += data
                data = self._sock.recv(4096)
        return result

    def send_command(self, cmd):
        self._send_command(cmd)
        size = int(self._sock.recv(4), 16)
        resp = self._sock.recv(size)
        # logger.debug('recv: %r' % resp[:200])
        self._sock.close()
        self._sock = None
        return resp.decode('utf8')

    def _transport(self, device_id):
        self._send_command('host:transport:%s' % device_id)

    def devices(self):
        '''adb devices
        '''
        result = self.send_command('host:devices')
        return result

    def shell(self, device_id, cmd, **kwds):
        '''adb shell
        '''
        cmd_line = 'shell:%s' % cmd
        self._transport(device_id)
        self._send_command(cmd_line)
        result = ADBPopen(self._sock, timeout=kwds['timeout']).communicate()
        self._sock = None
        return result

    def _sync_read_mode(self, remote_path):
        '''
        '''
        data = b'STAT' + struct.pack('I', len(remote_path)) + remote_path.encode('utf8')
        self._sock.send(data)
        result = self._sock.recv(16)
        if result[:4] != b'STAT':
            raise AdbError('sync_read_mode error')
        mode, size, time = struct.unpack('III', result[4:])
        return mode, size, time

    def pull(self, device_id, src_file, dst_file):
        '''adb pull
        '''
        time0 = time.time()
        self._transport(device_id)
        self._send_command('sync:')
        mode, fsize, ftime = self._sync_read_mode(src_file)
        if mode == 0:
            self._sock.close()
            self._sock = None
            raise AdbError('remote object %r does not exist' % src_file)
        data = b'RECV' + struct.pack('I', len(src_file)) + src_file.encode('utf8')
        self._sock.send(data)
        f = open(dst_file, 'wb')
        data_size = 0
        last_data = b''
        while True:
            result = self._sock.recv(8)
            if len(result) != 8:
                logger.warn('返回数据错误：%r' % result)
                
            last_data += result
            if len(last_data) < 8: 
                continue
            else: 
                result = last_data[:8]
                last_data = last_data[8:]

            psize = struct.unpack('I', result[4:])[0]  # 每个分包大小

            if result[:4] == b'DONE': break
            elif result[:4] == b'FAIL':
                raise AdbError(self._sock.recv(psize))
            elif result[:4] != b'DATA':
                raise AdbError('pull_file error')

            result = self._recv(psize - len(last_data))
            result = last_data + result
            if len(result) >= psize:
                last_data = result[psize:]
                result = result[:psize]
            else:
                raise ValueError('数据长度不一致,期望值:%d 实际值:%d' % (psize, len(result)))
            f.write(result)
            data_size += len(result)
            
        f.close()
        self._sock.send(b'QUIT' + struct.pack('I', 0))
        time_cost = time.time() - time0
        self._sock.close()
        self._sock = None
        if data_size > 0:
            return '%d KB/s (%d bytes in %fs)' % (int(data_size / 1000 / time_cost) if time_cost > 0 else 65535, data_size, time_cost)
        else:
            return ''
        
    def push(self, device_id, src_file, dst_file):
        '''adb push
        '''
        time0 = time.time()
        try:
            st = os.stat(src_file)
        except OSError as e:
            if e.errno == 2:
                raise AdbError("cannot stat '%s': No such file or directory" % src_file)
            else: raise e
        self._transport(device_id)
        self._send_command('sync:')
        mode, fsize, ftime = self._sync_read_mode(dst_file)
        
        s = b'%s,%d' % (dst_file.encode('utf8'), st.st_mode)
        data = b'SEND' + struct.pack('I', len(s)) + s
        self._sock.send(data)
        with open(src_file, 'rb') as fp:
            data = fp.read(SYNC_DATA_MAX)
            data_size = 0
            while data:
                send_data = b'DATA' + struct.pack('I', len(data)) + data
                self._sock.send(send_data)  
                data_size += len(data)
                data = fp.read(SYNC_DATA_MAX)

        data = b'DONE' + struct.pack('I', int(st.st_mtime))
        self._sock.send(data)
        result = self._sock.recv(8)
        if result[:4] == b'OKAY':
            self._sock.close()
            self._sock = None
            time_cost = time.time() - time0
            return '%d KB/s (%d bytes in %fs)' % (int(data_size / 1000.0 / time_cost) if time_cost > 0 else 0, data_size, time_cost)
        elif result[:4] == b'FAIL':
            msg_len = struct.unpack('I', result[4:])[0]
            error_msg = self._sock.recv(msg_len)
            self._sock.close()
            self._sock = None
            raise AdbError(error_msg)
        else:
            self._sock.close()
            self._sock = None
            raise RuntimeError('Unexpect data: %r' % result)
        
    def install(self, device_id, apk_path, args='', **kwds):
        '''adb install
        '''
        if not os.path.exists(apk_path):
            raise AdbError(r'can\'t find %r to install' % apk_path)
        apk_name = os.path.split(apk_path)[-1]
        dst_path = '/data/local/tmp/%s' % apk_name
        self.push(device_id, apk_path, dst_path)
        cmdline = 'pm install ' + (args + ' ' if args else '') + dst_path
        result = self.shell(device_id, cmdline, **kwds)
        return result[0].decode('utf8')

    def uninstall(self, device_id, package_name, **kwds):
        '''adb uninstall
        '''
        cmd = 'pm uninstall %s' % package_name
        result = self.shell(device_id, cmd, **kwds)
        return result[0].decode('utf8')

    def forward(self, device_id, local, remote):
        '''adb forward
        '''
        self._send_command('host-serial:%s:forward:%s;%s' % (device_id, local, remote))
        self._sock.close()
        self._sock = None
        return ''
    
    def remove_forward(self, device_id, local):
        '''adb forward --remove
        '''
        self._send_command('host-serial:%s:killforward:%s' % (device_id, local))
        self._sock.close()
        self._sock = None
        return ''
    
    def create_tunnel(self, device_id, remote_addr):
        '''创建与手机中服务端的连接通道
        '''
        self._transport(device_id)
        self._sock.settimeout(2)
        try:
            self._send_command(remote_addr)
        except AdbError as e:
            if 'closed' == e.args[0]:
                return ''
            raise
        except socket.timeout as e:
            logger.warn('create_tunnel timeout')
            return ''
        sock = self._sock
        self._sock = None
        return sock
        
    def get_state(self, device_id):
        '''获取设备状态
        '''
        return self.send_command('host-serial:%s:get-state' % (device_id))
    
    def connect(self, device_id):
        '''连接设备端口
        '''
        result = self.send_command('host:connect:%s' % device_id)
        return 'connected to' in result
    
    def disconnect(self, device_id):
        '''断开设备连接
        '''
        result = self.send_command('host:disconnect:%s' % device_id)
        return 'disconnected' in result
        
    def reboot(self, device_id, **kwds):
        '''重启设备
        '''
        self._transport(device_id)
        self._sock.settimeout(kwds['timeout'])
        try:
            self.send_command('reboot:')
        except socket.error as e:
            raise e
        except:
            pass
        return True
    
    def wait_for_device(self, device_id, **kwds):
        '''等待设备
        '''
        self._send_command('host-serial:%s:wait-for-any' % (device_id))
        return ADBPopen(self._sock, timeout=kwds['timeout']).communicate()
    
    def snapshot_screen(self, device_id):
        '''截屏
        
        return: Image.Image
        '''
        from PIL import Image
        self._transport(device_id)
        self._send_command('framebuffer:')
        
        fb_desc = self._sock.recv(13 * 4)
        version = struct.unpack_from('I', fb_desc, 0)[0]
        bpp = struct.unpack_from('I', fb_desc, 4)[0]
        size = struct.unpack_from('I', fb_desc, 8)[0]
        width = struct.unpack_from('I', fb_desc, 12)[0]
        height = struct.unpack_from('I', fb_desc, 16)[0]
        red_offset = struct.unpack_from('I', fb_desc, 20)[0]
        red_length = struct.unpack_from('I', fb_desc, 24)[0]  # @UnusedVariable
        blue_offset = struct.unpack_from('I', fb_desc, 28)[0]
        blue_length = struct.unpack_from('I', fb_desc, 32)[0]  # @UnusedVariable
        green_offset = struct.unpack_from('I', fb_desc, 36)[0]
        green_length = struct.unpack_from('I', fb_desc, 40)[0]  # @UnusedVariable
        alpha_offset = struct.unpack_from('I', fb_desc, 44)[0]
        alpha_length = struct.unpack_from('I', fb_desc, 48)[0]

        if version != 1:
            raise AdbError("Unsupported version of framebuffer: %s" % version)
        # detect order
        util_map = { red_offset: 'R', blue_offset: 'B', green_offset: 'G'}
        keys = list(util_map.keys())
        keys.sort()
        raw_mode = ''.join([util_map[it] for it in keys])
        
        # detect mode
        if alpha_length and alpha_offset:
            mode = 'RGBA'
            if bpp != 32:
                raise AdbError("Unsupported RGBA mode, bpp is %s" % bpp)
            raw_mode += 'A'
            
        elif alpha_offset:
            mode = 'RGBX'
            if bpp != 32:
                raise AdbError("Unsupported RGBX mode, bpp is %s" % bpp)
            raw_mode += 'X'
            
        else:
            mode = 'RGB'
            if bpp == 16:
                raw_mode += ';16'
            elif bpp == 24:
                pass
            else:
                raise AdbError("Unsupported RGB mode, bpp is %s" % bpp)

        data = b''
        while len(data) < size:
            data += self._sock.recv(4096)
        self._sock.close()
        self._sock = None
        return Image.frombuffer(mode, (width, height), data, 'raw', raw_mode, 0, 1)

if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''Android设备驱动
'''

import json
import re
import os
import time
import six
from qt4a.androiddriver.adb import ADB
from qt4a.androiddriver.clientsocket import AndroidSpyClient
from qt4a.androiddriver.util import SocketError, TimeoutError, KeyCode, logger

qt4a_path = '/data/local/tmp/qt4a'

def root_required(func):
    '''root权限装饰器
    '''
    def _wrap_func(self, *args, **kwargs):
        '''
        '''
        if not self.adb.is_rooted():
            raise RuntimeError('method %s need root device' % func.__name__)
        return func(self, *args, **kwargs)
    return _wrap_func

class DeviceDriver(object):
    '''Android设备驱动
    '''
    service_name = 'com.test.androidspy'
    service_port = 19862  # 部分机器只能使用TCP端口
    
    def __init__(self, adb):
        self._adb = adb
        self._device_id = adb.device_name
        self._is_local_device = ADB.is_local_device(self._device_id)
        self._client = None
        self._server_pid = 0
        self._timeout = 40

    def get_device_id(self):
        '''获取设备ID
        '''
        if not self._device_id:
            self._device_id = self.adb.run_shell_cmd('getprop ro.serialno')
        return self._device_id

    def run_driver_cmd(self, cmd, *args, **kwargs):
        '''执行驱动命令
        
        :param cmd:  命令
        :type  cmd:  string
        '''
        args = [('"%s"' % (it.replace('"', r'\"') if isinstance(it, str) else it)) for it in args]
        return self.adb.run_shell_cmd('sh %s/SpyHelper.sh %s %s' % (qt4a_path, cmd, ' '.join(args)), **kwargs)

    def get_language(self):
        '''获取系统语言
        '''
        return self.run_driver_cmd('getLanguage')

    def get_country(self):
        '''获取国家
        '''
        return self.run_driver_cmd('getCountry')

    @property
    def adb(self):
        return self._adb

    @property
    def client(self):
        if self._client == None:
            # 运行SpyHelper.sh
            self._client = self.run_server()
            # self._client.pre_connect()
        return self._client
    
    def get_device_imei(self):
        '''获取设备imei号
        '''
        if self.adb.is_rooted():
            return self.run_driver_cmd('getDeviceImei', root=True)
        else:
            return self._send_command('GetDeviceImei')
                
    def is_package_installed(self, pkg_name, pkg_size, pkg_md5):
        '''安装包是否已安装
        
        :param pkg_name:  应用包名
        :type pkg_name:   string
        :param pkg_size:  安装包大小
        :type pkg_size:   int
        :param pkg_md5:   安装包md5
        :type pkg_md5:    string
        '''
        ret = self.run_driver_cmd('isPackageInstalled', pkg_name, pkg_size, pkg_md5, root=self.adb.is_rooted())
        return 'true' in ret
    
    def install_package(self, pkg_path, overwrite=False):
        '''安装应用
        '''
        from qt4a.androiddriver.util import get_file_md5
        if not os.path.exists(pkg_path):
            raise RuntimeError('APK: %r not exist' % pkg_path)
        
        pkg_size = os.path.getsize(pkg_path)
        pkg_md5 = get_file_md5(pkg_path)
        pkg_name = ADB._get_package_name(pkg_path)
        if self.is_package_installed(pkg_name, pkg_size, pkg_md5):
            logger.info('APP %s [%d]%s is installed' % (pkg_name, pkg_size, pkg_md5))
            return True
        
        self.adb.install_apk(pkg_path, overwrite)
        return True
    
    def kill_process(self, package_name):
        '''杀死进程
        '''
        try:
            self.run_driver_cmd('killProcess', package_name, root=self.adb.is_rooted())
        except RuntimeError as e:
            logger.warn('killProcess error: %r' % e)
        if not self.adb.is_rooted(): return
        for _ in range(3):
            ret = self.adb.kill_process(package_name)
            if ret == None: return True
            elif ret == False: return False
        return False

    def get_external_sdcard_path(self):
        '''获取外置SD卡路径
        '''
        return self.run_driver_cmd('getExternalStorageDirectory', root=self.adb.is_rooted())

    def refresh_media_store(self, file_path=''):
        '''刷新图库，显示新拷贝到手机的图片
        '''
        from qt4a.androiddriver.adb import TimeoutError
        command = ''
        if not file_path:
            if hasattr(self, '_last_refresh_all') and time.time() - self._last_refresh_all < 2 * 60:
                logger.warn('[DeviceDriver] 120S内不允许全盘重复刷新')
                return
            sdcard_path = self.get_external_sdcard_path()
            command = 'am broadcast -a android.intent.action.MEDIA_MOUNTED --ez read-only false -d file://%s' % sdcard_path
            self._last_refresh_all = time.time()
        else:
            if file_path.startswith('/sdcard/'):
                file_path = self.get_external_sdcard_path() + file_path[7:]
            command = 'am broadcast -a android.intent.action.MEDIA_SCANNER_SCAN_FILE -d file://%s' % file_path
        try:
            self.adb.run_shell_cmd(command, self.adb.is_rooted())
        except TimeoutError:
            logger.exception('refresh_media_store failed')
            return False
        # self.adb.run_shell_cmd('am broadcast -a android.intent.action.BOOT_COMPLETED')
        if file_path and self.adb.get_sdk_version() >= 16:
            return self._wait_for_media_available(file_path)
    
    def _get_current_activity(self, has_package_name=False):
        '''获取当前Activity
        '''
        if has_package_name: has_package_name = 'true'
        else: has_package_name = 'false'
        ret = self.run_driver_cmd('getCurrentActivity', has_package_name, root=self.adb.is_rooted())
        return ret.strip().split('\n')[-1]
    
    def _get_current_window(self):
        '''使用dumpsys命令获取当前窗口
        '''
        result = self.adb.run_shell_cmd('dumpsys window windows')
        if result:
            line_list = result.split('\n')
            for line in line_list:
                if 'mCurrentFocus' in line:
                    result = line
                    break
        pattern = re.compile(r'mCurrentFocus=Window{(.+)}')
        ret = pattern.search(result)
        if not ret:
            logger.info('Get current window by dumpsys failed: %s' % result)
            return None
        result = ret.group(1).split(' ')[-1]
        if '/' in result: result = result.split('/')[-1]
        if 'Application Not Responding' in ret.group(1):
            result = 'Application Not Responding: %s' % result
        return result
    
    def get_current_activity(self):
        '''获取当前窗口
        '''
        timeout = 10
        time0 = time.time()
        result = None
        while time.time() - time0 < timeout:
            if not self.adb.is_rooted():
                result = self._get_current_window()
            else:
                try:
                    result = self._send_command('GetCurrentWindow')
                except SocketError as e:
                    raise e
                except RuntimeError as e:
                    logger.warn('GetCurrentWindow error: %s' % e)
            if not result or result == 'null':
                # 一般是设备黑屏之类
                time.sleep(0.5)
                continue
            return result
        
        if not self.adb.is_rooted(): return None
        
        logger.warn('GetCurrentWindow failed')
        return self._send_command('GetCurrentActivity')

    def get_screen_size(self):
        '''获取屏幕大小
        '''
        if self.adb.is_rooted():
            result = self._send_command('GetScreenSize')  # 使用Socket方式可能会出现获取到结果一直为0的情况
            if isinstance(result, dict) and result['Width'] > 0 and result['Height'] > 0:
                return result['Width'], result['Height']
            
        result = self.run_driver_cmd('getScreenSize', root=self.adb.is_rooted())
        result = result.split('\n')[-1]
        width, height = result.split(',')
        return int(width), int(height)
    
    def take_screen_shot(self, path, quality=90):
        '''截屏
        '''
        result = self.adb.run_shell_cmd('%s/screenshot capture -q %s' % (qt4a_path, quality), binary_output=True)
        # 为避免pull文件耗时，直接写到stdout
        if len(result) < 256:
            logger.warn('Take screenshot failed: %s' % result)
            return False
        with open(path, 'wb') as fp:
            fp.write(result)
        return True

    def drag(self, x1, y1, x2, y2, count=5, wait_time=40, send_down_event=True, send_up_event=True):
        '''在屏幕上拖拽
        '''
        if self.adb.is_rooted():
            self.client.send_command('Drag', X1=x1, Y1=y1, X2=x2, Y2=y2, StepCount=count, SleepTime=wait_time, SendDownEvent=send_down_event, SendUpEvent=send_up_event)
        else:
            # TODO: 改为访问QT4A助手服务进程
            self.run_driver_cmd('drag', x1, y1, x2, y2, count, wait_time, send_down_event, send_up_event)
            
    def _kill_server(self):
        '''杀死Server进程，用于Server卡死时
        '''
        if self.adb.is_rooted():
            return self.adb.kill_process(self.service_name + ':service')
        else:
            return self.adb.stop_service(self.service_name + '/.service.HelperService')
        
    def _server_opend(self):
        '''判断测试桩进程是否运行
        '''
        process_name = self.service_name + ':service'
        pid = self.adb.get_pid(process_name)
        if pid != 0 and pid != self._server_pid:
            logger.info('[DeviceDriver] %s pid is %d' % (process_name, pid))
        self._server_pid = pid
        return pid > 0

    def _run_server(self, server_name):
        '''运行系统测试桩
        '''
        if not self.adb.is_rooted():
            try:
                self.adb.start_service('%s/.service.HelperService' % server_name, {'serviceName': server_name})
            except RuntimeError:
                logger.warn('start helper server failed')
                self.adb.start_activity('%s/.activity.StartServiceActivity' % server_name, extra={'serviceName': 'HelperService'}, wait=False)
                time.sleep(1)
            return self._server_opend()
        
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            try:
                ret = self.run_driver_cmd('runServer', server_name, root=self.adb.is_rooted(), retry_count=1, timeout=10)
                logger.debug('Run server %s' % ret)
                if 'service run success' in ret:
                    # wait for parent process exit
                    time.sleep(0.5)
                elif 'service is running' in ret:
                    pass
                elif 'java.lang.UnsatisfiedLinkError' in ret:
                    raise RuntimeError('启动系统测试桩进程失败：\n%s' % ret)
            except TimeoutError as e:
                logger.warn('Run server timeout: %s' % e)
            
            if self._server_opend(): return True
            time.sleep(1)
        return False

    def _create_client(self, addr, port):
        '''创建Client实例
        '''
        return AndroidSpyClient(port, addr=addr, enable_log=False, timeout=self._timeout)
    
    def run_server(self, server_name=''):
        '''运行测试桩进程,创建服务端
        '''
        if server_name == '': server_name = self.service_name
        from qt4a.androiddriver.androiddriver import AndroidDriver
        port = AndroidDriver.get_process_name_hash(server_name, self.get_device_id())
        logger.info('[DeviceDriver] port is %d' % port)
        addr = '127.0.0.1'
        if not self._is_local_device: addr = self.adb.host_name
        if self._client: self._client.close()
        self._client = None
 
        time0 = time.time()
        timeout = 20
        kill_server = False
        
        server_type = 'localabstract'
        if self.adb.is_rooted() and self.adb.is_selinux_opened():
            # 创建TCP服务端
            server_name = str(self.service_port)
            server_type = 'tcp'

        while time.time() - time0 < timeout:
            if not kill_server and time.time() - time0 >= timeout // 2:
                # server进程存在问题，强杀
                self._kill_server()
                if self._client: self._client.close()
                self._client = None
                kill_server = True
            ret = self._run_server(server_name)
            logger.debug('[DeviceDriver] Server %s process created: %s' % (server_name, ret))
            
            while port < 65536:
                self._client = self._create_client(addr, port)
                port_is_opened = AndroidSpyClient.server_opened(port)
                if not port_is_opened: 
                    new_port = self.adb.forward(port, server_name, server_type)
                    if new_port != port:
                        self._client = self._create_client(addr, new_port)
                        logger.info('[AndroidDevice] new port=%d' % new_port)
                        port = new_port

                try:
                    hello_rsp = self.hello()
                    if hello_rsp == None: 
                        if port_is_opened:
                            # 需要释放端口
                            ret = self.adb.remove_forward(port)
                            logger.info('[AndroidDevice] remove %d forward: %s' % (port, ret))
                        break
                except RuntimeError:
                    port += 1
                    logger.info('[AndroidDevice] new port is %d' % port)
                    continue
                
                return self._client
            
        raise RuntimeError('连接系统测试桩超时')
    
    def _restart_server(self):
        '''重启系统测试桩
        '''
        if self._client:
            self._client.close()
            self._client = None
        self._kill_server()
        self.run_server()
        
    def _send_command(self, cmd_type, **kwds):
        '''发送命令
        '''
        result = self.client.send_command(cmd_type, **kwds)
        if result == None:
            logger.error('系统测试桩连接错误')
            self._kill_server()
            if self._client: self._client.close()
            self._client = None
            return self._send_command(cmd_type, **kwds)
            # raise SocketError('Socket连接错误')
        if 'Error' in result:
            raise RuntimeError(result['Error'])
        if not 'Result' in result:
            raise RuntimeError('%s返回结果错误：%s' % (cmd_type, result))
        return result['Result']

    def hello(self):
        '''
        '''
        result = self._client.hello()
        if result == None: return
        if not 'Result' in result:
            logger.warn('[DeviceDriver] no Result in hello rsp')
            raise RuntimeError('Server error')
        logger.info('[DeviceDriver] %s' % result['Result'])
        items = result['Result'].split(':')

        if len(items) > 0 and items[-1].isdigit():
            if self._server_pid == 0:
                logger.warn('[DeviceDriver] Server pid is 0')
            elif self._server_pid != 0 and int(items[-1]) != self._server_pid:
                logger.warn('[DeviceDriver] Server pid not match: %s' % (int(items[-1])))
                raise RuntimeError('Server error %s' % result['Result'])
        return result
    
    def close(self):
        if self._client != None:
            self._client.send_command('Exit')
            self._client = None

    def reboot(self, wait_cpu_low=True, usage=20, duration=10, timeout=120):
        '''重启手机
        
        :param wait_cpu_low:   是否等待CPU使用率降低
        :type wait_cpu_low:    bool
        :param usage:          cpu使用率阈值
        :type  usage:          int
        :param duration:       持续时间(秒)
        :type  duration:       int
        :param timeout:        超时时间，超市时间到后，无论当前CPU使用率是多少，都会返回
        :type  timeout:        int
        '''
        pattern = re.compile(r'^.+:\d+$')
        try:
            ret = self.run_driver_cmd('reboot', True, retry_count=1, timeout=30)
            if ret == 'false': return
        except RuntimeError as e:
            logger.warn('reboot: %r' % e)
            if pattern.match(self.device_id):
                try:
                    self.adb.reboot(60)
                except RuntimeError:
                    logger.warn('reboot: %s' % e)
                    # 虚拟机会出现明明重启成功却一直不返回的情况
            else:
                self.adb.reboot(0)  # 不能使用reboot shell命令，有些手机需要root权限才能执行

        time.sleep(10)  # 防止设备尚未关闭，一般重启不可能在10秒内完成
        self.adb.wait_for_boot_complete()
        self._adb = None  # 重启后部分属性可能发生变化,需要重新实例化

        if wait_cpu_low == True:
            self.wait_for_cpu_usage_low(usage, duration, timeout)

    def wait_for_cpu_usage_low(self, usage=20, duration=10, timeout=120):
        '''等待CPU使用率持续时间内在指定值以下
        :param usage:    cpu使用率阈值
        :type  usage:    int
        :param duration: 持续时间(秒)
        :type  duration: int
        :param timeout:  超时时间，超市时间到后，无论当前CPU使用率是多少，都会返回
        :type  timeout:  int
        '''
        time0 = time.time()
        time1 = time0
        while time.time() - time0 < timeout:
            cpu_usage = self.adb.get_cpu_usage()
            # print (cpu_usage)
            if cpu_usage > usage :
                time1 = time.time()
            if time.time() - time1 > duration:
                break
            time.sleep(1)

    def connect_wifi(self, wifi_name, wifi_pass=''):
        '''连接指定的Wifi
        '''
        if not self.adb.is_rooted():
            return self._send_command('ConnectWifi', WifiName=wifi_name)
        else:
            result = self.run_driver_cmd('connectWifi', wifi_name, wifi_pass, timeout=60)
            if not 'true' in result:
                logger.debug('Connect wifi result: %s' % result)
                return False
            return True
        
    def disable_wifi(self):
        '''禁用Wifi
        '''
        return self._send_command('DisableWifi')

    def enable_data_connection(self):
        '''启用数据连接
        '''
        result = self.run_driver_cmd('setDataConnection', 'true', root=self.adb.is_rooted(), timeout=60)
        if not 'true' in result:
            sim_state = self.get_sim_card_state()
            if 'SIM_STATE_ABSENT' in sim_state or 'SIM_STATE_UNKNOWN' in sim_state:
                raise RuntimeError('设备中没有SIM卡')
            raise RuntimeError('启用数据连接失败:%r' % result)
        return True

    def disable_data_connection(self):
        '''禁用数据连接
        '''
        result = self.run_driver_cmd('setDataConnection', 'false', root=self.adb.is_rooted(), timeout=60)
        if not 'true' in result:
            raise RuntimeError('禁用数据连接失败: %r' % result)
        return True

    def get_clipboard_text(self):
        '''获取剪切板内容
        '''
        result = self.run_driver_cmd('getClipboardText', root=self.adb.is_rooted())
        if six.PY2:
            return result.decode('raw-unicode-escape')
        return result
    
    def set_clipboard_text(self, text):
        '''设置剪贴板内容
        '''
        if six.PY2:
            text = text.encode('raw-unicode-escape')
        self.run_driver_cmd('setClipboardText', text, root=self.adb.is_rooted())
    
    @root_required
    def is_screen_lock_enabled(self):
        '''屏幕锁是否可用
        6.0以下需要系统权限：You either need MANAGE_USERS or CREATE_USERS permission to: query users
        '''
        return 'true' in self.run_driver_cmd('isScreenLockEnabled', root=True)
    
    @root_required
    def set_screen_lock_enable(self, enable=True):
        '''设置屏幕锁可用/不可用
        '''
        if enable: enable = 'true'
        else: enable = 'false'
        return 'true' in self.run_driver_cmd('setScreenLockEnable', enable, root=True)
        
    def is_screen_on(self):
        '''屏幕是否点亮
        '''
        return 'true' in self.run_driver_cmd('isScreenOn', root=self.adb.is_rooted())
    
    def wake_screen(self, wake=True):
        '''唤醒屏幕
        '''
        if self.is_screen_on() == wake: return True
        if self.adb.is_rooted():
            self.send_key(KeyCode.KEYCODE_POWER)
        else:
            return 'true' in self.run_driver_cmd('wakeScreen', 'true' if wake else 'false')
        return True
    
    def is_keyguard_locked(self):
        '''屏幕锁是否锁定
        '''
        return 'true' in self.run_driver_cmd('isKeyguardLocked', root=self.adb.is_rooted())
    
    def _lock_keyguard(self):
        '''
        '''
        return self._send_command('LockKeyguard')
    
    def _unlock_keyguard(self):
        '''
        '''
        return self._send_command('UnlockKeyguard')
    
    def lock_keyguard(self):
        '''锁屏
        '''
        if self.is_keyguard_locked(): return True  # 已经是锁屏状态
        if self.adb.is_rooted():
            self.set_screen_lock_enable(True)
        if self.adb.get_sdk_version() >= 16:
            # 发送电源键
            self.send_key(KeyCode.KEYCODE_POWER)
            return True
        
        self._last_activity_before_lock = self._get_current_activity(True)  # 保存锁屏前的Activity
        logger.debug('锁屏前Activity为：%r' % self._last_activity_before_lock)
        ret = self._lock_keyguard()
        if not ret: 
            logger.warn('lock keyguard failed')
            self.wake_screen(False)
            time.sleep(1)
            self.wake_screen(True)
            return self.is_keyguard_locked()
        return True
    
    def _unlock_keyguard_ge_16(self):
        '''4.1以上使用APP方式解锁
        '''
        self.adb.start_activity('com.test.androidspy/.activity.UnlockKeyguardActivity', wait=False)
        time.sleep(1)
        return True
        
    def unlock_keyguard(self):
        '''解锁屏幕
        '''
        if self.adb.is_rooted():
            if not self.is_keyguard_locked(): return True
            self.set_screen_lock_enable(False)
        if not self.is_keyguard_locked(): return True
        
        if self.adb.get_sdk_version() >= 16:
            self._unlock_keyguard_ge_16()
        else:
            # 使用接口解锁
            if self._unlock_keyguard(): return True
            old_activity = self._get_current_activity(True)
            logger.debug('before send HOME is %s' % old_activity)
            self.send_key(KeyCode.KEYCODE_HOME)

        return True

    def send_key(self, keys):
        '''发送按键
        '''
        if isinstance(keys, str):
            for item in KeyCode.get_key_list(keys):
                self.send_key(item)
            return
        elif isinstance(keys, list):
            # 组合键
            keys = [str(item) for item in keys]
            keys = ','.join(keys)
            
        if self.adb.is_rooted():
            logger.debug('SendKey %s' % keys)
            return self._send_command('SendKey', Keys=keys)
        else:
            return 'true' in self.run_driver_cmd('sendKey', keys)

    def get_mac_address(self):
        '''获取设备mac地址
        '''
        mac = self.run_driver_cmd('getWlanMac')
        if mac != 'null': return mac.replace(':', '')
        return None

    def _query_media_info(self, file_path):
        '''查询媒体文件信息
        '''
        result = self.run_driver_cmd('getMediaUri', file_path, root=self.adb.is_rooted())
        logger.debug('Get media uri result: %s' % result)
        if not result or 'content://' not in result:
            return ''
        return result[result.find('content://'):]

    def _wait_for_media_available(self, file_path, timeout=180):
        '''等待媒体文件存储到数据库
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            uri = self._query_media_info(file_path)
            if uri: return uri
            time.sleep(5)
        raise RuntimeError('等待媒体文件:%r 刷新超时' % file_path)

    def get_string_resource(self, pkg_name, string_id, lang=''):
        '''获取字符串资源
        '''
        result = self._send_command('GetStringResource', PkgName=pkg_name, Id=string_id, Lang=lang)
        return result

    def get_string_resource_id(self, pkg_name, text, lang=''):
        '''获取字符串资源ID
        '''
        result = self._send_command('GetStringResourceId', PkgName=pkg_name, Text=text, Lang=lang)
        if not result:
            if pkg_name == 'android': return {}
            return self.get_string_resource_id('android', text, lang)  # 查找系统资源
        return result

    def set_default_language(self, lang):
        '''设置默认语言
        '''
        lang_dict = {'中文': 'zh',
                     '英文': 'en',
                     '阿拉伯文': 'ar',
                     '德文': 'de',
                     '西班牙文': 'es',
                     '法文': 'fr',
                     '意大利文': 'it',
                     '日文': 'ja',
                     '韩文': 'ko'
                     }  # 语言对应关系
        country_dict = {'中国': 'CN',
                        '香港': 'HK',
                        '台湾': 'TW',
                        '美国': 'US',
                        '英国': 'GB',
                        '澳大利亚': 'AU',
                        '加拿大': 'CA',
                        '爱尔兰': 'IE',
                        '印度': 'IN',
                        '新西兰': 'NZ',
                        '新加坡': 'SG',
                        '南非': 'ZA',
                        '意大利': 'IT',
                        '瑞士': 'CH',
                        '日本': 'JP',
                        '韩国': 'KR',
                        }  # 国家(地区)对应关系
        
        pattern_lang = re.compile(r'^[a-z]{2}_[A-Z]{2}$')
        if not pattern_lang.match(lang):
            pattern = re.compile(r'(.+)\((.+)\)')
            if lang == '简体中文':
                lang = '中文(中国)'
            elif lang == '繁体中文':
                lang = '中文(台湾)'
            mat = pattern.match(lang)
            if mat:
                lang = '%s_%s' % (lang_dict[mat.group(1)], country_dict[mat.group(2)])
            else:
                # 没有地区
                lang = lang_dict[lang]
        self.run_driver_cmd('updateLangConfig', lang, root=self.adb.is_rooted())
        # 重启测试桩进程，保证重新加载资源
        self._restart_server()

    def get_static_field_value(self, pkg_name, cls_name, field_name, field_type):
        '''获取类中静态变量的值
        
        :param pkg_name:   包名
        :type pkg_name:    string
        :param cls_name:   类名
        :type cls_name:    string
        :param field_name: 字段名
        :type field_name:  string 
        '''
        return self.run_driver_cmd('getStaticFieldValue', pkg_name, cls_name, field_name, field_type)

    def get_system_boot_time(self):
        '''获取系统启动时间，单位为秒
        '''
        for _ in range(3):
            ret = self.run_driver_cmd('getSystemBootTime', root=self.adb.is_rooted())
            if ret and ret.isdigit(): return int(ret / 1000)
            logger.warn('getSystemBootTime return %r' % ret)
            time.sleep(2)
        return 0

    def get_available_storage(self, path):
        '''获取分区可用存储空间
        '''
        return int(self.run_driver_cmd('getAvailableStorage', path, root=self.adb.is_rooted()))

    @root_required
    def get_default_app(self, action):
        '''获取默认App
        '''
        return self.run_driver_cmd('getPreferedApp', action, root=True)
    
    @root_required
    def set_default_app(self, action, type, new_app):
        '''设置默认应用
        
        :param action: 应用针对的类型，如：android.media.action.IMAGE_CAPTURE
        :type action:  String
        :param new_app:新的应用包名
        :type new_app: String
        '''
        ret = self.run_driver_cmd('setPreferedApp', action, type, new_app, root=True)
        if 'false' in ret:
            raise RuntimeError('设置默认应用：%s 失败，检查应用是否安装并具有处理：%s 操作的能力' % (new_app, action))
        return ret
    
    @root_required
    def clear_default_app(self, action, type=''):
        '''清除默认应用设置
        需要android.permission.SET_PREFERRED_APPLICATIONS权限，只有系统应用可以获取
        '''
        return 'true' in self.run_driver_cmd('clearPreferedApp', action, type, root=True)

    def has_gps(self):
        '''是否有GPS
        '''
        return 'true' in self.run_driver_cmd('hasGPS')
    
    def get_camera_number(self):
        '''获取摄像头数目
        '''
        return int(self.run_driver_cmd('getCameraNumber', root=self.adb.is_rooted()))

    def is_debug_package(self, package_name):
        '''是否是debug包
        '''
        ret = self.run_driver_cmd('isDebugPackage', package_name, root=self.adb.is_rooted())
        logger.info('isDebugPackage ret: %s' % ret)
        if 'NameNotFoundException' in ret:
            raise RuntimeError('APP: %s not installed' % package_name)
        return 'true' in ret
    
    def get_view_id(self, package_name, view_str_id):
        '''获取控件整型ID
        '''
        id_list = []  # 可能会存在多个整型ID（应用和android）
        for package_name in [package_name, 'android']:
            time0 = time.time()
            result = self._send_command('GetViewId', PkgName=package_name, StrId=view_str_id)
            if result > 0:
                logger.debug('get_view_id %s in %s use %s S, result=%d' % (view_str_id, package_name, time.time() - time0, result))
                id_list.append(result)
            else:
                if package_name == 'android' and len(id_list) == 0:
                    raise RuntimeError('View id: %s not exist' % view_str_id)
        return id_list
    
    def get_resource_origin_name(self, package_name, res_type, confuse_name):
        '''获取资源原始名称
        '''
        time0 = time.time()
        result = self._send_command('GetResourceOriginName', PkgName=package_name, ResType=res_type, ConfuseName=confuse_name)
        logger.debug('get_resource_origin_name %s in %s use %s S, result=%s' % (confuse_name, package_name, time.time() - time0, result))
        return result
    
    def start_activity(self, params):
        '''使用系统测试桩启动Activity
        '''
        if isinstance(params['FileUri'], (list, tuple)):  # and len(param['FileUri']) > 10:
            # 使用Socket方式，命令行方式有参数长度限制
            result = self._send_command('StartActivity', **params)
            if not result in range(5):
                raise RuntimeError('Start activity %s failed: %s' % (activity_name, result))
        else:
            result = self.run_driver_cmd('startActivity', json.dumps(param), root=self.adb.is_rooted())
            if 'Error:' in result or 'Exception' in result:
                raise RuntimeError('Start activity %s failed: %s' % (activity_name, result))
    
    def play_sound(self, file_path):
        '''播放语音
        
        :param file_path: 音频文件在手机中的路径
        '''
        return self._send_command('PlaySound', FilePath=file_path)
    
    def set_volume(self, volume):
        '''设置音量
        '''
        self.run_driver_cmd('setVolume', int(volume), root=self.adb.is_rooted())
        
    def get_phone_contacts(self):
        '''获取手机联系人列表
        '''
        result = self._send_command('GetPhoneContacts')
        return json.loads(result)
        
    def add_phone_contacts(self, name, phone):
        '''添加手机联系人
        '''
        return 'true' in self._content_provider_patch_func(self.run_driver_cmd)('addPhoneContacts', name, phone, root=self.adb.is_rooted())
    
    def del_phone_contacts(self, name):
        '''删除手机联系人
        '''
        return 'true' in self._content_provider_patch_func(self.run_driver_cmd)('delPhoneContacts', name, root=self.adb.is_rooted())

    def _content_provider_patch_func(self, func):
        '''解决4.0以下版本手机中访问ContentProvider的限制
        '''
        def wrap_func(*args, **kwargs):
            if self.adb.get_sdk_version() >= 16:
                return func(*args, **kwargs)
            else:
                result = func(*args, **kwargs)
                if 'java.lang.SecurityException' in result:
                    for _ in range(3):
                        ret = self.adb.run_shell_cmd('.%s/inject system_server' % qt4a_path, True, timeout=60, retry_count=1)
                        if 'Inject Success' in ret: break
                    result = func(*args, **kwargs)
                return result
        return wrap_func

    def is_shortcut_exist(self, package_name):
        '''判断快捷方式是否存在
        '''
        result = self._content_provider_patch_func(self.run_driver_cmd)('isAppShortcutExist', package_name, root=self.adb.is_rooted())
        return 'true' in result
    
    def set_app_permission(self, package_name, perm_name, is_allowed=True):
        '''设置应用权限
        '''
        if self.adb.get_sdk_version() < 19: return False
        result = self.run_driver_cmd('setAppPermission', package_name, perm_name, 'true' if is_allowed else 'false', root=self.adb.is_rooted())
        return 'true' in result
    
    def modify_system_setting(self, type, name, value):
        '''修改系统设置
        
        :param type: 设置类型，一般为system
        :type type:  string
        :param name: 要设置的键值
        :type name:  string
        :param value:要设置的数值
        :type value: string/int
        '''
        result = self._content_provider_patch_func(self.run_driver_cmd)('modifySystemSetting', type, name, value, root=self.adb.is_rooted())
        return 'true' in result
    
    def is_vpn_connected(self):
        '''VPN是否连接
        '''
        result = self.run_driver_cmd('isVpnConnected')
        return 'true' in result
    
    def connect_vpn(self, vpn_type, server, username, password):
        '''连接VPN
        
        :param vpn_type:  VPN类型
        :type vpn_type:   int
        :param server:    服务器地址
        :type server:     string
        :param username:  用户名
        :type username:   string
        :param password:  密码
        :type password:   string
        '''
        result = self.run_driver_cmd('connectVpn', vpn_type, server, username, password, root=self.adb.is_rooted())
        return 'true' in result
    
    def disconnect_vpn(self):
        '''断开VPN
        '''
        result = self.run_driver_cmd('disconnectVpn', root=self.adb.is_rooted())
        return 'true' in result
    
    def end_call(self):
        '''挂断电话
        '''
        result = self.run_driver_cmd('endCall', root=self.adb.is_rooted())
        return 'true' in result
    
    def grant_all_runtime_permissions(self, package_name):
        '''给APP授予所有运行时权限
        '''
        result = self.run_driver_cmd('grantAllRuntimePermissions', package_name, root=self.adb.is_rooted())
        return not 'null' in result

    def resolve_domain(self, domain):
        '''解析域名
        '''
        result = self.run_driver_cmd('resolveDomain', domain, root=self.adb.is_rooted())
        return result.strip()

    def set_radio_enabled(self, enable):
        '''是否启用Radio
        '''
        return 'true' in self.run_driver_cmd('setRadioEnabled', 'true' if enable else 'false', root=self.adb.is_rooted())
    
    def set_http_proxy(self, host, port):
        '''设置http代理
        '''
        return 'true' in self.run_driver_cmd('setHttpProxy', host if host else '', port if port else '', root=self.adb.is_rooted())
    
    def connect_screenshot_service(self):
        '''连接截屏服务
        '''
        timeout = 10
        time0 = time.time()
        screenshot_service_name = 'qt4a_screenshot'
        while time.time() - time0 < timeout:
            sock = self.adb.create_tunnel(screenshot_service_name, 'localabstract')
            if sock: return sock
            self.adb.run_shell_cmd('%s/screenshot runserver -d' % qt4a_path)
        else:
            raise RuntimeError('start screenshot service failed')
    
    def unzip_file(self, zip_file, save_dir):
        '''解压文件
        
        :param zip_file: 压缩文件路径
        :type  zip_file: string
        :param save_dir: 保存目录路径
        :type  save_dir: string
        '''
        result = self.run_driver_cmd('unzip', zip_file, save_dir)
        return 'true' in result

if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''Android端浏览器封装
'''

import re
import time

from qt4a.andrcontrols import Window, TextView, WebView
from qt4a.androidapp import AndroidApp
from qt4a.device import Device
from qt4a.qpath import QPath
from qt4w.browser import IBrowser
from qt4w.webcontrols import WebPage

class QT4ABrowser(AndroidApp, IBrowser):
    '''QT4A助手实现的浏览器功能
    '''
    package_name = 'com.test.androidspy'
    process_name = package_name + ':browser'
    
    def __init__(self, device=None):
        if not device:
            from testbase import context
            tc = context.current_testcase()
            if tc:
                device = tc.acquire_device()
            else:
                # 使用本地设备
                device = Device()
        super(QT4ABrowser, self).__init__(self.process_name, device)
        
    def open_url(self, url, page_cls=None):
        '''打开一个url，返回page_cls类的实例
        
        :param url: 要打开页面的url
        :type url:  string
        :param page_cls: 要返回的具体WebPage类,为None表示返回WebPage实例
        :type page_cls: Class
        '''
        self.device.start_activity(self.package_name + '/' + QT4ABrowserWindow.Activity, data_uri=url)
        if page_cls == None: page_cls = WebPage
        browser_window = QT4ABrowserWindow(self)
        webview = browser_window.Controls['WebView']
        return page_cls(webview)
        
    def find_by_url(self, url, page_cls=None, timeout=10):
        '''在当前打开的页面中查找指定url,返回WebPage实例，如果未找到，返回None
        
        :param url: 要查找的页面url
        :type url:  string
        :param page_cls: 要返回的具体WebPage类,为None表示返回WebPage实例
        :type page_cls: Class
        :param timeout: 查找超时时间，单位：秒
        :type timeout: int/float
        '''
        time0 = time.time()
        if page_cls == None: page_cls = WebPage
        pattern = re.compile(url)
        while time.time() - time0 < timeout:
            browser_window = QT4ABrowserWindow(self)
            webview = browser_window.Controls['WebView']
            page = page_cls(webview)
            page_url = page.url
            if page_url == url or pattern.match(page_url): return page
            time.sleep(1)
        else:
            raise RuntimeError('find url %s failed' % url)
        
class QT4ABrowserWindow(Window):
    '''QT4A浏览器窗口
    '''
    Activity = QT4ABrowser.package_name + '.activity.BrowserActivity'
    Process = QT4ABrowser.process_name
    
    def __init__(self, browser):
        super(QT4ABrowserWindow, self).__init__(browser)
        self.updateLocator({'标题': {'type': TextView, 'root': self, 'locator': QPath('/Id="text_title"')},
                            'WebView': {'type': WebView, 'root': self, 'locator': QPath('/Id="webView1"')},
                            })




# -*- coding:UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''管理命令
'''

from __future__ import print_function

import argparse
import os
import sys
import time

try:
    raw_input
except:
    raw_input = input
    
def install_qt4a_driver(args):
    from qt4a.androiddriver.adb import ADB
    from qt4a.androiddriver.androiddriver import copy_android_driver
    
    device_list = ADB.list_device()
    if len(device_list) == 0:
        raise RuntimeError('No Android device found')
    elif len(device_list) == 1:
        device_id = device_list[0]
    elif len(device_list) > 1:
        text = '\nCurrent Android device list:\n'
        for i, dev in enumerate(device_list):
            text += '%d. %s\n' % ((i + 1), dev)

        while True:
            print(text)
            result = raw_input('Please input the index of device to install driver:\n')
            if result.isdigit():
                if int(result) > len(device_list):
                    sys.stderr.write('\nIndex %s out of range\nValid index range: [1, %d]\n' % (result, len(device_list)))
                    time.sleep(0.1)
                    continue
                device_id = device_list[int(result) - 1]
            else:
                if not result in device_list:
                    sys.stderr.write('\nDevice id %r not exist\n' % result)
                    time.sleep(0.1)
                    continue
                device_id = result
            break
        
    print('Device "%s" will install driver...' % device_id)
    copy_android_driver(device_id, args.force)
    print('Install QT4A driver to %s completely.' % device_id)

def repack_apk(args):
    from qt4a.apktool.repack import repack_apk
    print('Repacking apk %s...' % (' '.join(args.path)))
    outpath = repack_apk(args.path, args.debuggable)
    print('Repack apk completely.\nOutput apk path is: ')
    if isinstance(outpath, list):
        for it in outpath:
            print(it)
    else:
        print(outpath)
    
def qt4a_manage_main():
    from qt4a.androiddriver.util import OutStream
    sys.stdout = OutStream(sys.stdout)
    sys.stderr = OutStream(sys.stderr)
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='subcommand')
    install_driver_parser = subparsers.add_parser('install-driver', help='install qt4a driver')
    install_driver_parser.add_argument('-f', '--force', action='store_true', help='force install qt4a driver')
    install_driver_parser.set_defaults(func=install_qt4a_driver)
    
    repack_parser = subparsers.add_parser('repack-apk', help='repack apk file')
    repack_parser.add_argument('-p', '--path', nargs='*', required=True, help='path of apks to repack')
    repack_parser.add_argument('-d', '--debuggable', type=bool, default=True, help='whether apk debuggable after repack')
    repack_parser.set_defaults(func=repack_apk)
    
    args = parser.parse_args()

    if hasattr(args, 'func'): 
        args.func(args)
    else:
        parser.print_help()
        print('\n%s: error: too few arguments' % os.path.split(sys.argv[0])[-1], file=sys.stderr) # show error info in python3
        
        
if __name__ == '__main__':
    qt4a_manage_main()
    
    
    

